/*
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * @OSKIT-FLUX-GPLUS@
 */
#include <stand.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>

#include "net.h"
#include "netif.h"

#include "emuboot.h"
#include "bootwhat.h"

static struct {
	u_char header[HEADER_SIZE];
	boot_info_t boot_info;
} rpkt;

static ssize_t 
recvbootinfo(struct iodesc *d, void *pkt, size_t len, time_t tleft)
{
	struct udphdr *uh;
	struct ip *ip;

	errno = 0;
	len = readudp(d, pkt, len, tleft);
	if (len == -1)
		return (-1);

	uh = (struct udphdr *)pkt - 1;
	ip = (struct ip *)uh - 1;
	if (ip->ip_src.s_addr != rootip.s_addr ||
	    uh->uh_sport != htons(BOOTWHAT_DSTPORT)) {
#ifdef DEBUG
		printf("bootinfo reply not from server (%s/%d != %s/%d)\n",
		       inet_ntoa(ip->ip_src), ntohs(uh->uh_sport),
		       inet_ntoa(rootip), BOOTWHAT_DSTPORT);
#endif
		return (-1);
	}

	/*
	 * Coming from our server and the right port,
	 * any other errors are fatal.
	 */
	if (len != sizeof(rpkt.boot_info) ||
	    rpkt.boot_info.opcode != BIOPCODE_BOOTWHAT_REPLY) {
#ifdef DEBUG
		printf("bootinfo bogus reply from %s/%d, len=%d, opcode=%d\n",
		       inet_ntoa(ip->ip_src), ntohs(uh->uh_sport),
		       len, rpkt.boot_info.opcode);
#endif
		errno = EINVAL;
		return (-1);
	}

	if (rpkt.boot_info.status != BISTAT_SUCCESS) {
		printf("No bootinfo for us on server %s\n",
		       inet_ntoa(rootip));
		errno = EINVAL;
		return (-1);
	}

	return (len);
}

int
bootinfo_getinfo(boot_what_t **boot_whatp)
{
	struct {
		u_char header[HEADER_SIZE];
		boot_info_t boot_info;
	} pkt;
	struct iodesc *iodesc;
	int rv, fd;

	fd = open(NETDEV, 0);
	if (fd < 0) {
#ifdef DEBUG
		int err = errno;
		printf("bootinfo: could not open net device %s err=%d\n",
		       NETDEV, err);
		errno = err;
#endif
		return (errno);
	}

	iodesc = socktodesc(fd);
	if (iodesc == 0) {
		close(fd);
		return (ENODEV);
	}

	memset(&pkt.boot_info, 0, sizeof(pkt.boot_info));
	pkt.boot_info.opcode = BIOPCODE_BOOTWHAT_REQUEST;

	iodesc->myport = htons(BOOTWHAT_SRCPORT);
	iodesc->destip = rootip;
	iodesc->destport = htons(BOOTWHAT_DSTPORT);
	iodesc->xid = 1;
#ifdef DEBUG
	printf("bootinfo: request from %s/%d to ",
	       inet_ntoa(iodesc->myip), iodesc->myport);
	printf("%s/%d\n", inet_ntoa(iodesc->destip), iodesc->destport);
#endif

	rv = sendrecv(iodesc, sendudp, &pkt.boot_info, sizeof(pkt.boot_info),
		      recvbootinfo, &rpkt.boot_info, sizeof(rpkt.boot_info));
	close(fd);

	if (rv == -1) {
		if (errno == ETIMEDOUT)
			printf("No reply from bootinfo server %s\n",
			       inet_ntoa(rootip));
		return (errno);
	}

	/*
	 * Got a seemingly valid boot info packet that says what to do.
	 */
	*boot_whatp = (boot_what_t *) &rpkt.boot_info.data;
	return 0;
}

void
bootinfo_showinfo(void)
{
	boot_what_t *boot_whatp;

	if (bootinfo_getinfo(&boot_whatp) != 0)
		return;

	printf("Default boot:\n");
	switch (boot_whatp->type) {

	case BIBOOTWHAT_TYPE_MB:
		printf("boot multiboot kernel `%s:%s' with cmdline `%s'\n",
		       inet_ntoa(boot_whatp->what.mb.tftp_ip),
		       boot_whatp->what.mb.filename,
		       boot_whatp->cmdline);
		break;

	case BIBOOTWHAT_TYPE_PART:
		printf("boot from main disk, partition %d\n",
		       boot_whatp->what.partition);
		break;

	case BIBOOTWHAT_TYPE_SYSID:
		printf("boot from main disk partition with sysid %d\n",
		       boot_whatp->what.sysid);
		break;

	default:
		printf("!? invalid action %d\n", boot_whatp->type);
		break;
	}
}
