/*
 * diskboot.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#include <stand.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/udp.h>

#include "btxv86.h"
#include "libi386/libi386.h"
#include "emuboot.h"
#include "bootwhat.h"
#include "bootstrap.h"

/*
 * Support for disk booting
 */

static void *magicaddr = (void *)0x7C00;

struct bios_mbr {
	char bootcode[446];
	char partinfo[64];	/* 4 x bios_partition_info */
	unsigned short magic;
};

/* magic */
#define BIOS_MAGIC	0xaa55

struct bios_partition_info {
        unsigned char bootid;
        unsigned char beghead;
        unsigned char begsect;
        unsigned char begcyl; 
        unsigned char sysid;
        unsigned char endhead;
        unsigned char endsect;
        unsigned char endcyl;
        unsigned long offset;
        unsigned long numsectors;
};

/* bootable (active) bootid */
#define BIOS_BOOTABLE	0x80

static struct bios_mbr diskmbr;
static char partbr[512];
static int load_mbr(int dieonerror);
static int check_mbr(void);
static void show_mbr(void);
static int load_sector(int sectno, void *buf);

void
disk_showinfo(void)
{
	if (load_mbr(0) == 0) {
		printf("Partitions, %s\n", DISKDEV);
		show_mbr();
	}
}

/*
 * Load the Master Boot Record using the BIOS.
 */
static int
load_mbr(int dieonerror)
{
	if (load_sector(0, &diskmbr) == 0 && check_mbr() == 0)
		return 0;

	if (dieonerror)
		panic("Could not load/verify MBR");
	return -1;
}

/*
 * Check sanity of the MBR.
 */
static int
check_mbr(void)
{
	struct bios_mbr *mbr;
	struct bios_partition_info *entry;
	int i;

	mbr = &diskmbr;
#ifdef DEBUG
	printf("MBR@%p: magic=%x, partinfo@%p\n", mbr, mbr->magic,
	       mbr->partinfo);
#endif
	if (mbr->magic != BIOS_MAGIC) {
#ifdef DEBUG
		printf("MBR has bad magic (%x != %x)\n",
		       mbr->magic, BIOS_MAGIC);
#endif
		return -1;
	}

	entry = (struct bios_partition_info *) mbr->partinfo;
	for (i = 0; i < 4; i++, entry++) {
#ifdef DEBUG
		printf("%c%d: sysid=%d, offset=%lu, nsect=%lu\n",
		       entry->bootid==BIOS_BOOTABLE ? '*' : ' ', i+1,
		       entry->sysid, entry->offset, entry->numsectors);
#endif
		if (entry->numsectors == 0)
			continue;

		/* XXX sanity check fields */
	}

	return 0;
}

static struct ptypes {
	int sysid;
	char *str;
} sysids[] = {
	{ 165,	"BSD" },
	{ 131,	"Linux" },
	{ 130,	"Linux swap" },
	{ 7,	"NTFS" },
	{ 0,	"Unused" },
	{ -1,	"Unknown" }
};

static char *
ptypestr(int sysid)
{
	struct ptypes *pt = sysids;

	for (pt = sysids; pt->sysid >= 0; pt++)
		if (pt->sysid == sysid)
			break;
	return (pt->str);
}

static void
show_mbr(void)
{
	struct bios_partition_info *entry;
	int i;

	entry = (struct bios_partition_info *) diskmbr.partinfo;
	for (i = 0; i < 4; i++, entry++) {
		if (entry->numsectors == 0)
			continue;
		printf("%c%d: sysid=%d (%s), offset=%lu, nsect=%lu\n",
		       entry->bootid==BIOS_BOOTABLE ? '*' : ' ', i+1,
		       entry->sysid, ptypestr(entry->sysid),
		       entry->offset, entry->numsectors);
	}
}

/*
 * Copy the given sector contents into the magic load address,
 * switch to 16bit/real mode, and then jump to the address.
 *
 * Note that we copy the sector into the memory previously occupied
 * by the pxeldr, this way we should not be clobbering anything.
 */
static void
boot_from(void *buf)
{
	void __restart(void *) __attribute__((__noreturn__));

	/*
	 * Shutdown PXE
	 */
	dev_cleanup();

	i386_copyin(buf, (vm_offset_t)magicaddr, 512);

	/*
	 * FreeBSD's mbr, boot0 and boot1 all expect %dl to be set to
	 * the boot device (where 0x80 == HD0) when they are invoked.
	 * I don't know if this is a BIOS handoff feature or if it is
	 * a bug and everyone assumes it is a bit of PC lore.
	 *
	 * Anyway, we set it here and it makes it through the BTX intact.
	 */
	__asm __volatile("movb $0x80,%dl");

	__restart(magicaddr);
}

/*
 * Boot from the MBR.  As long as there is a MBR, and its intact,
 * something good should happen (say, Lilo will appear).
 */
static void
boot_mbr(void)
{
	boot_from(&diskmbr);
}

/*
 * Load and boot the first sector of the indicated partition.
 */
static void
boot_partition(int pnum, struct bios_partition_info *ptab)
{
	if (load_sector((int)ptab->offset, partbr) != 0 ||
	    ((struct bios_mbr *)partbr)->magic != BIOS_MAGIC)
		panic("Could not load/verify partition %d boot record", pnum);

	boot_from(partbr);
}

/*
 * Simply load and boot the MBR right off the disk.
 */
void
boot_mbr_default(void)
{
	printf("Loading MBR and booting active partition\n");

	(void) load_mbr(1);
	boot_mbr();
}

/*
 * Boot from a partition with a particular sysid.
 * Search the table for the first non-empty partition of the given type.
 * If the type does not exist, return non-zero.
 */
int
boot_mbr_sysid(int sysid)
{
	struct bios_partition_info *entry;
	int partition;

	printf("Loading MBR and booting system ID %d\n", sysid);

	(void) load_mbr(1);

	/*
	 * Look for a non-empty partition that matches the one we want.
	 */
	entry = (struct bios_partition_info *)diskmbr.partinfo;
	for (partition = 0; partition < 4; partition++, entry++) {
		if (entry->numsectors == 0)
			continue;
		if (entry->sysid == sysid)
			break;
	}
	if (partition == 4) {
		printf("%d: sysid not found in any partition\n", sysid);
		return 1;
	}

	if (set_active(partition+1) != 0)
		printf("WARNING: could not set active partition in MBR\n");
	boot_partition(partition+1, entry);
	return 0;
}

/*
 * Boot from a particular partition.
 * If the partition does not exist, return non-zero.
 */ 
int
boot_mbr_partition(int partition)
{
	struct bios_partition_info *entry;

	printf("Loading MBR and booting system partition %d\n", partition);

	if (partition < 1 || partition > 4) {
		printf("%d: invalid partition number\n", partition);
		return 1;
	}

	/* partition is 1-based, table is 0-based */
	partition--;

	(void) load_mbr(1);

	entry = ((struct bios_partition_info *)diskmbr.partinfo) + partition;
	if (entry->numsectors == 0) {
		printf("%d: non-existent partition\n", partition+1);
		return 1;
	}

	if (set_active(partition+1) != 0)
		printf("WARNING: could not set active partition in MBR\n");
	boot_partition(partition+1, entry);
	return 0;
}

static int
load_sector(int sectno, void *buf)
{
	int fd, rv = 0;

	fd = open(DISKDEV, 0);
	if (fd >= 0) {
		if (lseek(fd, (off_t)sectno * 512, SEEK_SET) < 0 ||
		    read(fd, buf, 512) < 512) {
			rv = errno;
#ifdef DEBUG
			printf("short/failed read of sector %d, err=%x\n",
			       sectno, rv);
#endif
		}
		close(fd);
	} else {
		rv = errno;
#ifdef DEBUG
		printf("failed open of %s, err=%x\n", DISKDEV, rv);
#endif
	}

#ifdef DEBUG
	printf("Got sector %d: magic=0x%x\n",
	       sectno, ((unsigned short *)buf)[255]);
#endif
	return rv;
}

#include <btxv86.h>

/*
 * Set the active partition in the disk MBR
 */
int
set_active(int apart)
{
	struct bios_partition_info *entry;
	int i;

	if (apart < 1 || apart > 4) {
		printf("set_active: invalid partition number %d\n", apart);
		return 1;
	}

	if (load_mbr(0) != 0) {
		printf("set_active: could not load MBR\n");
		return 1;
	}

	/*
	 * Already active
	 */
	entry = &((struct bios_partition_info *)diskmbr.partinfo)[apart-1];
	if (entry->bootid & BIOS_BOOTABLE)
		return 0;

	/*
	 * Verify the desired partition and mark it as active
	 */
	if (entry->numsectors == 0) {
		printf("set_active: emptry partition %d\n", apart);
		return 1;
	}
	entry->bootid |= BIOS_BOOTABLE;

	/*
	 * Clear any other active indicator.
	 */
	apart--;
	entry = (struct bios_partition_info *)diskmbr.partinfo;
	for (i = 0; i < 4; i++, entry++)
		if (i != apart && (entry->bootid & BIOS_BOOTABLE))
			entry->bootid &= ~BIOS_BOOTABLE;

	/*
	 * Write MBR back out.
	 *
	 * XXX we do this totally behind the back of the boot loader
	 * subsystem (biosdisk) since it doesn't support disk writes.
	 * Thus, this is the last thing you should do before jumping
	 * to the next level boot code.
	 *
	 * This code lifted from the biosdisk read code.
	 */
	v86.ctl = V86_FLAGS;
	v86.addr = 0x13;
	v86.eax = 0x300 | 1;
	v86.ecx = 1;
	v86.edx = 0x80;		/* XXX: HDD0 */
	v86.es = VTOPSEG(&diskmbr);
	v86.ebx = VTOPOFF(&diskmbr);
	v86int();
	if (v86.efl & 0x1) {
		printf("set_active: could not write back MBR, eax=0x%x\n",
		       v86.eax);
		return 1;
	}
	return 0;
}
