/*-
 * Copyright (c) 1998, 2002 Michael Smith <msmith@freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD: src/sys/boot/i386/loader/main.c,v 1.17.2.5 2000/12/28 13:12:51 ps Exp $
 */

//#define RUNWHOAMI

/*
 * MD bootstrap main() and assorted miscellaneous
 * commands.
 */

#include <stand.h>
#include <string.h>
#include <machine/bootinfo.h>
#include <sys/reboot.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/udp.h>

#include "net.h"
#include "bootstrap.h"
#include "libi386/libi386.h"
#include "btxv86.h"

#include "emuboot.h"
#include "bootwhat.h"

#define	KARGS_FLAGS_CD		0x1
#define	KARGS_FLAGS_PXE		0x2

/* Arguments passed in from the boot1/boot2 loader */
static struct 
{
    u_int32_t	howto;
    u_int32_t	bootdev;
    u_int32_t	bootflags;
    u_int32_t	pxeinfo;
    u_int32_t	res2;
    u_int32_t	bootinfo;
} *kargs;

static u_int32_t	initial_howto;
static u_int32_t	initial_bootdev;
static struct bootinfo	*initial_bootinfo;

struct arch_switch	archsw;		/* MI/MD interface boundary */

static void		extract_currdev(void);
static int		isa_inb(int port);
static void		isa_outb(int port, int value);
void			exit(int code);

/* from vers.c */
extern	char bootprog_name[], bootprog_rev[], bootprog_date[], bootprog_maker[];

/* XXX debugging */
extern char end[];

int
main(void)
{
    int i;
    int err;
    boot_what_t *boot_whatp;

    /* Pick up arguments */
    kargs = (void *)__args;
    initial_howto = kargs->howto;
    initial_bootdev = kargs->bootdev;
    initial_bootinfo = kargs->bootinfo ? (struct bootinfo *)PTOV(kargs->bootinfo) : NULL;

    /* 
     * Initialise the heap as early as possible.  Once this is done, malloc() is usable.
     */
    bios_getmem();

    setheap((void *)end, (void *)bios_basemem);

    /* 
     * XXX Chicken-and-egg problem; we want to have console output early, but some
     * console attributes may depend on reading from eg. the boot device, which we
     * can't do yet.
     *
     * We can use printf() etc. once this is done.
     * If the previous boot stage has requested a serial console, prefer that.
     */
    if (initial_howto & RB_SERIAL)
	setenv("console", "comconsole", 1);
    if (initial_howto & RB_MUTE)
	setenv("console", "nullconsole", 1);
    cons_probe();

    /*
     * Initialise the block cache
     */
    bcache_init(32, 512);	/* 16k cache XXX tune this */

    /*
     * We only want the PXE disk to try to init itself in the below walk through
     * devsw if we actually booted off of PXE.
     */
    if((kargs->bootinfo == NULL) &&
       ((kargs->bootflags & KARGS_FLAGS_PXE) != 0)) {
	pxe_enable(kargs->pxeinfo ? PTOV(kargs->pxeinfo) : NULL);
    }

    /*
     * March through the device switch probing for things.
     */
    for (i = 0; devsw[i] != NULL; i++)
	if (devsw[i]->dv_init != NULL)
	    (devsw[i]->dv_init)();

    printf("\n%s, Revision %s\n", bootprog_name, bootprog_rev);
    printf("(%s, %s)\n", bootprog_maker, bootprog_date);

    extract_currdev();

    archsw.arch_autoload = i386_autoload;
    archsw.arch_getdev = i386_getdev;
    archsw.arch_copyin = i386_copyin;
    archsw.arch_copyout = i386_copyout;
    archsw.arch_readin = i386_readin;
    archsw.arch_isainb = isa_inb;
    archsw.arch_isaoutb = isa_outb;

    {
	char cip[16], sip[16], gip[16];

	/* XXX force DHCP query */
	close(open(NETDEV, 0));

	strcpy(cip, inet_ntoa(myip));
	strcpy(sip, inet_ntoa(rootip));
	strcpy(gip, inet_ntoa(gateip));

	printf("DHCP Info: IP: %s, Server: %s, Gateway: %s\n",
	       cip, sip, gip);
    }

    /*
     * Give the user a few seconds to interrupt and go into
     * interactive mode. If we return, go through the auto
     * boot and hope it works.
     */
    printf("Type a key for interactive mode (quick, quick!)\n");
    for (i = 0; i < 3000; i++) {
	    if (ischar() == 0)
		    delay(1000);
	    else {
		    getchar();
		    boot_interactive();
		    break;
	    }
    }

    /*
     * See what the server thinks we should do.
     * If no response or an error, boot the default partition.
     */
    err = bootinfo_getinfo(&boot_whatp);
    if (err != 0)
	    goto mbr;

    switch (boot_whatp->type) {

    case BIBOOTWHAT_TYPE_MB:
	err = load_multiboot_image(&boot_whatp->what.mb.tftp_ip,
				   boot_whatp->what.mb.filename,
				   boot_whatp->cmdline);
	break;

    case BIBOOTWHAT_TYPE_PART:
	err = boot_mbr_partition(boot_whatp->what.partition);
	break;

    case BIBOOTWHAT_TYPE_SYSID:
	err = boot_mbr_sysid(boot_whatp->what.sysid);
	break;

    default:
	panic("bootinfo invalid action %d\n", boot_whatp->type);
	break;
    }

    /*
     * Well, the only way to get here is if the specified boot failed.
     * Fall through to booting the default, and hope it works.
     */
 mbr:
#ifdef RUNWHOAMI
    /*
     * First try running the whoami kernel which prints out info about
     * the node.  Used at node setup time.
     */
    load_multiboot_image(0, "/tftpboot/whoami", "");
#endif
    printf("Server specified boot failed (err=%d), "
	   "booting from default partition...\n", err);

    boot_mbr_default();
    exit(0);
}

/*
 * Allow for an interactive boot specification. Returning implies that
 * we go through the automated boot.
 */
void
boot_interactive(void)
{
#define PROMPT	"Enter boot spec: "
	char	buf[512];
	
	while (1) {
		/*
		 * Flush!
		 */
		while (ischar() != 0)
			getchar();

		printf("%s", PROMPT);
		gets(buf);

#define PART	"part:"
		if (strncmp(buf, PART, strlen(PART)) == 0) {
			int part = strtol(&buf[strlen(PART)], 0, 10);
			boot_mbr_partition(part);			
		}
#define SYSID	"sysid:"
		else if (strncmp(buf, SYSID, strlen(SYSID)) == 0) {
			int sysid = strtol(&buf[strlen(SYSID)], 0, 10);
			boot_mbr_sysid(sysid);
		}
#define AUTO	"auto"
		else if (strncmp(buf, AUTO, strlen(AUTO)) == 0) {
			return;
		}
#define REBOOT	"reboot"
		else if (strncmp(buf, REBOOT, strlen(REBOOT)) == 0) {
			printf("Rebooting ...\n");
			exit(0);
		}
#define INFO	"info"
		else if (strncmp(buf, INFO, strlen(INFO)) == 0) {
			disk_showinfo();
			bootinfo_showinfo();
			continue;
		}
#define ACTIVE	"active:"
		else if (strncmp(buf, ACTIVE, strlen(ACTIVE)) == 0) {
			int active = strtol(&buf[strlen(ACTIVE)], 0, 10);
			if (set_active(active) == 0)
				printf("active partition now %d\n", active);
			continue;
		}
		else if (isalpha(buf[0]) || isdigit(buf[0]) || buf[0] == '/') {
			struct		in_addr	ip;
			char		*path, *args;
			
			parse_multiboot_spec(buf, &ip, &path, &args);
			load_multiboot_image(&ip, path, args);
		}

		printf("Choices are: \n"
		       "  part:<partition number>, or\n"
		       "  sysid:<system ID>, or\n"
		       "  <pathway to multiboot image>, or\n"
		       "  auto (query the bootinfo server again), or\n"
		       "  info (show boot options), or\n"
		       "  active:<partition number>, or\n"
		       "  reboot\n");
	}
}

/*
 * Set the 'current device' by (if possible) recovering the boot device as 
 * supplied by the initial bootstrap.
 *
 * XXX should be extended for netbooting.
 */
static void
extract_currdev(void)
{
    struct i386_devdesc	new_currdev;
    int			major, biosdev;

    /* Assume we are booting from a BIOS disk by default */
    new_currdev.d_dev = &biosdisk;
    new_currdev.d_type = new_currdev.d_dev->dv_type;

    /* new-style boot loaders such as pxeldr and cdldr */
    if (kargs->bootinfo == NULL) {
        if ((kargs->bootflags & KARGS_FLAGS_CD) != 0) {
	    /* we are booting from a CD with cdldr */
	    new_currdev.d_kind.biosdisk.slice = -1;
	    new_currdev.d_kind.biosdisk.partition = 0;
	    biosdev = initial_bootdev;
	} else if ((kargs->bootflags & KARGS_FLAGS_PXE) != 0) {
	    /* we are booting from pxeldr */
	    new_currdev.d_dev = &pxedisk;
	    new_currdev.d_type = new_currdev.d_dev->dv_type;
	    new_currdev.d_kind.netif.unit = 0;
	    biosdev = -1;
	} else {
	    /* we don't know what our boot device is */
	    new_currdev.d_kind.biosdisk.slice = -1;
	    new_currdev.d_kind.biosdisk.partition = 0;
	    biosdev = -1;
	}
    } else if ((initial_bootdev & B_MAGICMASK) != B_DEVMAGIC) {
	/* The passed-in boot device is bad */
	new_currdev.d_kind.biosdisk.slice = -1;
	new_currdev.d_kind.biosdisk.partition = 0;
	biosdev = -1;
    } else {
	new_currdev.d_kind.biosdisk.slice = (B_ADAPTOR(initial_bootdev) << 4) +
					     B_CONTROLLER(initial_bootdev) - 1;
	new_currdev.d_kind.biosdisk.partition = B_PARTITION(initial_bootdev);
	biosdev = initial_bootinfo->bi_bios_dev;
	major = B_TYPE(initial_bootdev);

	/*
	 * If we are booted by an old bootstrap, we have to guess at the BIOS
	 * unit number.  We will loose if there is more than one disk type
	 * and we are not booting from the lowest-numbered disk type 
	 * (ie. SCSI when IDE also exists).
	 */
	if ((biosdev == 0) && (B_TYPE(initial_bootdev) != 2))	/* biosdev doesn't match major */
	    biosdev = 0x80 + B_UNIT(initial_bootdev);		/* assume harddisk */
    }
    
    /*
     * If we are booting off of a BIOS disk and we didn't succeed in determining
     * which one we booted off of, just use disk0: as a reasonable default.
     */
    if ((new_currdev.d_type == devsw[0]->dv_type) &&
	((new_currdev.d_kind.biosdisk.unit = bd_bios2unit(biosdev)) == -1)) {
	printf("Can't work out which disk we are booting from.\n"
	       "Guessed BIOS device 0x%x not found by probes, defaulting to disk0:\n", biosdev);
	new_currdev.d_kind.biosdisk.unit = 0;
    }
    env_setenv("currdev", EV_VOLATILE, i386_fmtdev(&new_currdev),
	       i386_setcurrdev, env_nounset);
    env_setenv("loaddev", EV_VOLATILE, i386_fmtdev(&new_currdev), env_noset,
	       env_nounset);
}

/* provide this for panic, as it's not in the startup code */
void
exit(int code)
{
    __exit(code);
}

/* ISA bus access functions for PnP, derived from <machine/cpufunc.h> */
static int		
isa_inb(int port)
{
    u_char	data;
    
    if (__builtin_constant_p(port) && 
	(((port) & 0xffff) < 0x100) && 
	((port) < 0x10000)) {
	__asm __volatile("inb %1,%0" : "=a" (data) : "id" ((u_short)(port)));
    } else {
	__asm __volatile("inb %%dx,%0" : "=a" (data) : "d" (port));
    }
    return(data);
}

static void
isa_outb(int port, int value)
{
    u_char	al = value;
    
    if (__builtin_constant_p(port) && 
	(((port) & 0xffff) < 0x100) && 
	((port) < 0x10000)) {
	__asm __volatile("outb %0,%1" : : "a" (al), "id" ((u_short)(port)));
    } else {
        __asm __volatile("outb %0,%%dx" : : "a" (al), "d" (port));
    }
}
