/*
 * multiboot.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#include <stand.h>
#include <string.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/udp.h>
#include <sys/param.h>
#include <sys/linker.h>
#include <machine/elf.h>

#include "bootstrap.h"
#include "libi386/libi386.h"
#include "btxv86.h"
#include "net.h"

#include "emuboot.h"
#include "multiboot.h"

#define DEBUG

static struct multiboot_info   *boot_info;

static int init_boot_info(char *filename, char *cmdline);
static void multiboot_info_dump(struct multiboot_info *bi);
static void boot_start(struct multiboot_info *bi, vm_offset_t entry);

int
load_multiboot_image(struct in_addr *tftpserver, char *filename, char *cmdline)
{
	vm_offset_t entry = 0;
	int err;

#ifdef DEBUG
	printf("Creating Multiboot goo ...\n");
#endif
	err = init_boot_info(filename, cmdline);
	if (err) {
#ifdef DEBUG
		printf("Could not build multiboot header, err=%d\n", err);
#endif
		return err;
	}

	/*
	 * XXX tweak some magic global variables while we do the TFTP load.
	 */
	{
		struct loaded_module *mp = 0;
		struct module_metadata *md;
		struct in_addr Orootip;
		char Orootpath[FNAME_SIZE];

		Orootip.s_addr = rootip.s_addr;
		strncpy(Orootpath, rootpath, FNAME_SIZE);

		if (tftpserver && tftpserver->s_addr)
			rootip.s_addr = tftpserver->s_addr;
		rootpath[0] = '\0';

#ifdef DEBUG
		printf("Loading ELF file `%s' from %s...\n",
		       filename, inet_ntoa(rootip));
#endif
		err = elf_loadmodule(filename, 0, &mp);
		if (err == 0) {
			md = mod_findmetadata(mp, MODINFOMD_ELFHDR);
			if (md != 0)
				entry = ((Elf_Ehdr *)&(md->md_data))->e_entry
					& 0xffffff;
			else
				err = EINVAL;
		}
		strncpy(rootpath, Orootpath, FNAME_SIZE);
		rootip.s_addr = Orootip.s_addr;

		if (err) {
#ifdef DEBUG
			printf("Load failed, err=%d\n", err);
#endif
			if (mp)
				mod_discard(mp);
			return (err);
		}
#ifdef DEBUG
		printf("Loaded at [0x%x - 0x%x], entry=0x%x\n",
		       mp->m_addr, mp->m_addr + mp->m_size, entry);
#endif
	}

#ifdef DEBUG
	printf("Booting the kernel ...\n");
#endif
	__exec((void *)VTOP(boot_start), VTOP(boot_info), entry, 0);

	/*
	 * Never returns, but of course
	 */
	return 0;
}

static void
boot_start(struct multiboot_info *bi, vm_offset_t entry)
{
	asm volatile("call	*%%ecx" : : "b" (bi), "c" (entry));
}

/*
 * Take a string and separate the IP, path, and arguments into strings.
 */
int
parse_multiboot_spec(char *cline, struct in_addr *ip, char **path, char **args)
{
	char	*p;

	if (isdigit(*cline) && (p = strchr(cline, ':'))) {
		*p++ = '\0';
		ip->s_addr = inet_addr(cline);
	} else {
		ip->s_addr = 0;
		p = cline;
	}

	*path = p;
	
	if ((p = strchr(p, ' '))) {
		*p++ = '\0';
		*args = p;
	} else
		*args = "";

	return 0;
}

static int
init_boot_info(char *filename, char *cmdline)
{
	char *dest;

	/* Allocate memory for the boot_info structure and modules array.  */
	boot_info = (struct multiboot_info*)calloc(sizeof(*boot_info), 1);
	if (boot_info == 0)
		return (ENOMEM);

	boot_info->mods_count = 0;
	boot_info->mods_addr  = 0;

	/* Fill in the upper and lower memory size fields in the boot_info.  */
	boot_info->flags |= MULTIBOOT_MEMORY;

	/*
	 * Find lower and upper bounds of physical memory.
	 * We look in the NVRAM to find the size of base memory and
	 * use extended_mem_size() to find the size of extended memory.
	 */
	boot_info->mem_lower = bios_basemem / 1024;
	boot_info->mem_upper = bios_extmem / 1024;

	/*
	 * Boot arguments
	 */
	dest = calloc(2048, 1);
	if (dest == 0) {
		free(boot_info);
		return (ENOMEM);
	}
	boot_info->cmdline = VTOP(dest);
	boot_info->flags |= MULTIBOOT_CMDLINE;
	strcpy(dest, filename);
	strcat(dest, " ");

#ifdef FORCE_SERIAL_CONSOLE
	/* Force serial line console */
	strcat(dest, " -h -f ");
#endif
	if (!cmdline[0] || !strstr(cmdline, "--"))
		strcat(dest, " -- ");
	if (cmdline[0])
		strcat(dest, cmdline);

	printf("Kernel Command Line: %s\n", dest);

	/*
	 * - No command lines passed up from 16 bit mode.
	 * - No Boot module support.
	 *	The loaded kernel needs to be wrapped in a multiboot adaptor
	 *	if it wants boot modules.
	 */

#ifdef DEBUG
	multiboot_info_dump(boot_info);
#endif

	return (0);
}

static void
multiboot_info_dump(struct multiboot_info *bi)
{
	struct multiboot_module *m;
	unsigned i;

	printf("MultiBoot Info (flags: 0x%x)\n", bi->flags);

	if (bi->flags & MULTIBOOT_MEMORY)
		printf("  PC Memory: lower %dK, upper %dK\n",
		       bi->mem_lower, bi->mem_upper);

	if (bi->flags & MULTIBOOT_BOOT_DEVICE)
		printf("  Boot device: {%d,%d,%d,%d}\n",
		       bi->boot_device[0],
		       bi->boot_device[1],
		       bi->boot_device[2],
		       bi->boot_device[3]);

	if (bi->flags & MULTIBOOT_CMDLINE)
		printf("  Kernel command line: `%s'\n",
		       (char *)PTOV(bi->cmdline));

	if (bi->flags & MULTIBOOT_MODS) {
		printf("  Boot modules: %d\n", bi->mods_count);
		m = (struct multiboot_module *)PTOV(bi->mods_addr);
		for (i = 0; i < bi->mods_count; i++) {
			printf("  Module %d: %08x-%08x (size %d)\n",
			       i, m[i].mod_start, m[i].mod_end,
			       m[i].mod_end - m[i].mod_start);
			if (m[i].string)
				printf("    String: `%s' at %08x\n",
				       (char*)PTOV(m[i].string), m[i].string);
		}
	}

	if (bi->flags & MULTIBOOT_AOUT_SYMS)
		printf("  Symbol table (a.out): start 0x%08x,"
		       " symtab 0x%08x, strtab 0x%08x\n",
		       bi->syms.a.addr,
		       bi->syms.a.tabsize, bi->syms.a.strsize);

	if (bi->flags & MULTIBOOT_ELF_SHDR)
		/* XXX */
		printf("  Has MULTIBOOT_ELF_SHDR info\n");

	if (bi->flags & MULTIBOOT_MEM_MAP) {
		struct AddrRangeDesc *rdesc;
		unsigned end;

		printf("  Memory Map:\n      addr: 0x%08x,"
		       " count: %d\n", 
		       bi->mmap_addr, bi->mmap_count);

		/* XXX Count is currently the length in bytes of the
		 * entire structure: GRUB and the multiboot spec
		 * actually call it mmap_length. 
		 */
		rdesc = (struct AddrRangeDesc *)bi->mmap_addr;
		end = bi->mmap_addr + bi->mmap_count;

		while (end > (int)rdesc) {
			/* 
			 * I'd rather use .0 precision for the High
			 * entries (since they're guaranteed to be 0
			 * on a PC), but this works too.  */
			printf("      base: 0x");
			if (rdesc->BaseAddrHigh)
				printf("%08lx", rdesc->BaseAddrHigh);
			printf("%08lx length: 0x", rdesc->BaseAddrLow);

			if (rdesc->LengthHigh)
				printf("%08lx", rdesc->LengthHigh);
			printf("%08lx type: ", rdesc->LengthLow);

			switch (rdesc->Type) {
			case 1:
				printf("memory\n");
				break;
			case 2:
				printf("reserved\n");
				break;
			default:
				printf("undefined (%ld)\n", rdesc->Type);
			}

			(char *)rdesc += rdesc->size + 4;
		};
	}
	
	
}
