/*
 * lossfraction.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/proc.h>
#include <sys/jail.h>
#include <sys/queue.h>

#include <vm/vm_zone.h> /* memory for tfc_cb and the likes */

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_pcb.h>
#include <netinet/in_var.h>
#include <netinet/ip_var.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>

#include <netinet/xtcp/tfc.h>
#include <netinet/xtcp/tfci.h>
#include <netinet/xtcp/tfi.h>
#include <netinet/xtcp/tfc_stats.h>
#include <netinet/xtcp/tfc_error.h>
#include <netinet/xtcp/lossfraction.h>
#include <netinet/xtcp/scoreboard.h>
#include <netinet/xtcp/tfc_cb.h>

#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif

#define min(a, b) ((a) < (b) ? (a) : (b))
#define max(a, b) ((a) > (b) ? (a) : (b))

#define SCALE_FACTOR 1000000 /* 10^6 seems enough for now */
#define SCALE_ROOT 1000

#define FLOATIZE(n)     ((n)*SCALE_FACTOR)
#define DEFLOATIZE(n)   ((n)/SCALE_FACTOR)
#define SQRT_FL(n)   ((n)*SCALE_ROOT)
#define SQRT_DEFL(n)       ((n)/SCALE_ROOT)

#define MIN_MULT ((int)FLOATIZE(0.5))  /* lower bound for deweighting factor (new_mult) */

#define DW_FACTOR 2
/* last interval must be DW_FACTOR larger than avg. to start deweighting */

#define EQN_B 1

/* Let the compiler do some floating points for us */
static long long weights[WEIGHTS] = { FLOATIZE(1), FLOATIZE(1), FLOATIZE(1), FLOATIZE(1),
			      FLOATIZE(0.8), FLOATIZE(0.6), FLOATIZE(0.4), FLOATIZE(0.2)
                            };
static int nweights = WEIGHTS;

#define isqrt mborg_isqrt4

static unsigned long mborg_isqrt4(unsigned long val);

/* return sending rate as packets-per-rtt */
long tf_rate(long long p, long rtt, long rto)
{
	unsigned long long tmp1, tmp2, tmp3, tmp4, tmp5;
	unsigned long long rate;
	
	tmp1 = min(3 * isqrt((unsigned long) 3 * EQN_B * p / 8), SCALE_ROOT*1);
	/* root */
	
	tmp2 = rtt * isqrt(2 * EQN_B * p / 3);
	/* root */
	
	tmp3 = FLOATIZE(1) + DEFLOATIZE(32 * p * p);
	/* float */
	
	tmp5 = SQRT_DEFL(rto * tmp1 * p);
	/* float */
	
	tmp4 = tmp5 * tmp3;
	/* float * float */
	
	tmp4 = DEFLOATIZE(tmp4);
	tmp2 = SQRT_FL(tmp2);

	if ((tmp2+tmp4)) {
		rate = FLOATIZE(1)*SCALE_ROOT /(tmp2 + tmp4);
		/* calculate packet rate per rtt */
		rate *= rtt;
		rate = SQRT_DEFL(rate);
	} else
		rate = 0;

	/*
	  printf ("tmp1 %llu\n", tmp1);
	  printf ("tmp2 %llu\n", tmp2);
	  printf ("tmp3 %llu\n", tmp3);	
	  printf ("tmp4 %llu\n", tmp4);
	  printf ("tmp5 %llu\n", tmp5);
	*/
	/*return (long) (rate) ;*/

	/* AJW fudge factor */
	return (long) (rate * 6 / 4);
}

long long
fake_rate(long pktrate, long rtt, long t0)
{
	long long lf = lf_eq(pktrate, rtt, t0);
	if (lf == 0) {
		printf("loss fraction is 0\n");
		return FLOATIZE(1)/1000;
	}
	else 
	       return FLOATIZE(1)/lf;
}

/*
 * Do a sort of binary search to look for an appropriate loss rate
 */
long long
lf_eq(long pktrate, long rtt, long t0) {
        /* assume loss fraction is between 1/100000 and 1.0 */
        long long l1 = FLOATIZE(1), l2 = 0, l3 = FLOATIZE(1) * 1 / 100000;
        long long b1 = tf_rate(l1, rtt, t0), b2 = 0, b3 = tf_rate(l3, rtt, t0);
        int count = 100;

        if (b1 > pktrate)
                return l1;
        if (b3 < pktrate)
                return l3;

        while (--count) {
                l2 = l1 + (l3 - l1) / 2;
                b2 = tf_rate(l2, rtt, t0);
                if (b2 > pktrate + 8) {
                        l3 = l2;
                } else if (b2 < pktrate - 2) {
                        l1 = l2;
                } else {
                        break;
                }
        }
	printf ("Steady allowed rate = %lld...", b2);
        return l2;
}

long long
lossfraction(struct tfc_cb *tp)
{
	/* use mult to deweight old loss intervals */
        long long pplavg1 = 0, pplavg2 = 0;
        long long weightsum1 = 0, weightsum2 = 0, new_mult;
        int size1 = 0, size2 = 0, ppl1, ppl2;
        int num_intervals = 0;
	int ppl_cur = tp->ppl_open;
        int i;
	long long fraction = 0;

        /* loss avg. without latest interval */
        for (i = 0; i < nweights; ++i) {
                ppl2 = tp->ppl[(ppl_cur - i + nweights) % (nweights + 1)];
                if (ppl2 > 0) {
                        /* pplavg2 += ppl2 * weights[i] * mult[i]; */
                        /* weightsum2 += weights[i] * mult[i]; */
                        pplavg2 += ppl2 * weights[i];
                        weightsum2 += weights[i];
                        ++num_intervals;
                }
        }

	/* determine deweighting factor */
	if (num_intervals > 0 && weightsum2 > 0 && pplavg2 > 0
	    && tp->ppl[ppl_cur] > 0 && tp->ppl[ppl_cur] > 2 * pplavg2) {
		pplavg2 /= weightsum2;
		new_mult = min((DW_FACTOR * pplavg2) / tp->ppl[ppl_cur], FLOATIZE(1));
	} else {
		new_mult = FLOATIZE(1);
	}

	/* reset for second computation */
	pplavg2 = 0;
	weightsum2 = 0;
	
	/* compute lossrate with/without packets after last loss
	 * --> only take these packets into account if that decreases the loss rate */
	for (i = 0; i < nweights; ++i) {
		ppl1 = tp->ppl[(ppl_cur - i + nweights + 1) % (nweights + 1)];
		ppl2 = tp->ppl[(ppl_cur - i + nweights) % (nweights + 1)];
		/* no deweighting of current interval */
		if (ppl1 > 0) {
			if (i > 0) {
				long long m1 = max(new_mult, MIN_MULT);
				long long m2 = DEFLOATIZE(weights[i] * m1);
				weightsum1 += m2;
				pplavg1 += ppl1 * m2;
			} else {
				weightsum1 += weights[i];
				pplavg1 += ppl1 * weights[i];
			}
			size1 += ppl1;
		}
		if (ppl2 > 0) {
			if (i > 0) {
				long long m3 = max(new_mult, MIN_MULT);
				long long m4 = DEFLOATIZE(weights[i] * m3);
				weightsum2 += m4;
				pplavg2 += ppl2 * m4;
			} else {
				weightsum2 += weights[i];
				pplavg2 += ppl2 * weights[i];
			}
			size2 += ppl2;
		}
		/* printf("%lld \t %lld, %lld %lld \n", pplavg1, pplavg2, weightsum1,
		   weightsum2); */
	}
	/* printf("\n"); */

	if (pplavg1 > 0 ) {
		pplavg1 = (long long)pplavg1/weightsum1;
		pplavg1 = FLOATIZE(1) / pplavg1;
	}
	if (pplavg2 > 0 ) {
		pplavg2 = (long long)pplavg2/weightsum2;
		pplavg2 = FLOATIZE(1) / pplavg2;
	}
	fraction = min(pplavg1, pplavg2);
	
        return fraction;
}

/* Integer square root routine
   Taken from: http://www.azillionmonkeys.com/qed/sqroot.html
 */

static unsigned long
mborg_isqrt4(unsigned long val) {

	unsigned long temp, g=0;

	if (val >= 0x40000000) {
		g = 0x8000; 
		val -= 0x40000000;
	}
	
#define INNER_MBGSQRT(s)                      \
  temp = (g << (s)) + (1 << ((s) * 2 - 2));   \
  if (val >= temp) {                          \
    g += 1 << ((s)-1);                        \
    val -= temp;                              \
  }
	
	INNER_MBGSQRT (15)
	INNER_MBGSQRT (14)
	INNER_MBGSQRT (13)
	INNER_MBGSQRT (12)
	INNER_MBGSQRT (11)
	INNER_MBGSQRT (10)
	INNER_MBGSQRT ( 9)
	INNER_MBGSQRT ( 8)
        INNER_MBGSQRT ( 7)
        INNER_MBGSQRT ( 6)
        INNER_MBGSQRT ( 5)
        INNER_MBGSQRT ( 4)
        INNER_MBGSQRT ( 3)
        INNER_MBGSQRT ( 2)

#undef INNER_MBGSQRT

	temp = g+g+1;
	if (val >= temp) g++;
	return g;
}

