/*
 * scoreboard.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/proc.h>
#include <sys/jail.h>

#include <vm/vm_zone.h> /* memory for tfc_cb and the likes */

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_pcb.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>

#include <netinet/xtcp/tfc.h>
#include <netinet/xtcp/tfci.h>
#include <netinet/xtcp/tfi.h>
#include <netinet/xtcp/tfc_stats.h>
#include <netinet/xtcp/tfc_error.h>
#include <netinet/xtcp/lossfraction.h>
#include <netinet/xtcp/scoreboard.h>
#include <netinet/xtcp/tfc_cb.h>

static void update_rtt(tick_t new_rtt, tick_t ts, struct tfc_cb *tp); 
static void enter_slowstart(struct tfc_cb *tp, tick_t ts);

/* #define SB_TRACE */

int
sb_insert(seq_t n, u_char nonce, struct tfc_cb* tp, tick_t ts)
{
	/* Enter this packet in the scoreboard */
	int nlp1 = SB_INC(tp->sb_lastplus1);

	if (nlp1 == tp->sb_first) {
		/* Overflow */

		/* printf ("Scoreboard overflow for seq %d\n", n); */
		tfc_stats.sb_overflows++;

		/* XXX: Maybe cope with the overflow */
		return TFC_SBOF;
	}
	
	tp->scoreboard[tp->sb_lastplus1].seq_no = n;
	tp->scoreboard[tp->sb_lastplus1].nonce = nonce;
	tp->scoreboard[tp->sb_lastplus1].ts = ts;
	tp->sb_lastplus1 = nlp1;
	tp->scoreboard[tp->sb_lastplus1].state = SBSTATE_SENT;
	return 0;
}


int
sb_update_leq(seq_t seq_no, tick_t ts, struct tfc_cb *tp, u_char *nonce)
{
	int idx = sb_lookup_leq(seq_no, tp);
	if (idx < 0)
		return EINVAL;

	tp->scoreboard[idx].ts = ts;
	*nonce = tp->scoreboard[idx].nonce;

	/* change the sequence number  */
	tp->scoreboard[idx].seq_no = seq_no;

	return 0;
}

int
sb_update(seq_t seq_no, tick_t ts, struct tfc_cb *tp, u_char *nonce)
{
	int idx = sb_lookup(seq_no, tp);
	if (idx < 0)
		return EINVAL;

	tp->scoreboard[idx].ts = ts;
	*nonce = tp->scoreboard[idx].nonce;
	return 0;
}


int 
sb_matchsum_leq(int eseq, u_char nonce_sum, tick_t ts, struct tfc_cb *tp) {
        int error = 0;
	int bidx = tp->sb_first;
	int eidx;
	u_char sum = tp->nonce_sum;

	eidx = sb_lookup_leq(eseq, tp);
	if (eidx < 0)
		return TFC_EINVALSEQ;
			

	if (tp->nonce_sumsynched == 0) { /* && TFC_SEQ_GEQ(seq_no, tp->nonce_sumseq)) { */
	  /*printf("Nonce sum synched to %d at seq %d\n", nonce_sum, eseq); */
		tp->nonce_sumsynched = 1;
		goto out2;
	}
	
	/* Check that the sum is valid */

	while(bidx != eidx) {
		sum = NONCE_SUM(tp->scoreboard[bidx].nonce, sum);
		bidx = SB_INC(bidx);
	}


	/* The range is inclusive */
	sum = NONCE_SUM(tp->scoreboard[bidx].nonce, sum);

	if (sum != nonce_sum) {
		tfc_stats.badnonce++;
		error = TFC_BADNONCE;
		goto out2;
	}

	/* Match */
	/* printf("Nonce sum matched for seq %d\n", eseq); */
 out2:
	if (error == 0) {
		int acked = SB_INC(eidx);
		tick_t rtto = TICK_DIFF(ts, tp->scoreboard[bidx].ts);
		acked = SB_DIFF(acked, tp->sb_first);
		update_rtt(rtto, ts, tp);
		tp->nonce_sum = nonce_sum;
		tp->last_ack_seq = eseq;
		tp->last_ack_ts = tp->scoreboard[eidx].ts;
		tp->nonce_sumseq = eseq;
		tfc_stats.acktotal += acked;
		tp->ppl[tp->ppl_open] += acked; 
		tp->sb_first = SB_INC(eidx);
		tp->ack_cur += acked;
		/* tp->last_acked = eseq; */
	}
	return error;
}	



/* match sum, the range is inclusive of sb_first and eidx */
int
sb_matchsum(int eseq, u_char nonce_sum, tick_t ts, struct tfc_cb *tp)
{
	int error = 0;
	int bidx = tp->sb_first;
	int eidx;
	u_char sum = tp->nonce_sum;

	eidx = sb_lookup(eseq, tp);
	if (eidx < 0)
		return TFC_EINVALSEQ;
			
	if (tp->nonce_sumsynched == 0) { /* && TFC_SEQ_GEQ(seq_no, tp->nonce_sumseq)) { */
		/* printf("Nonce sum synched to %d at seq %d\n", nonce_sum, eseq); */
		tp->nonce_sumsynched = 1;
		goto out;
	}
	
	/* Check that the sum is valid */

	while(bidx != eidx) {
		sum = NONCE_SUM(tp->scoreboard[bidx].nonce, sum);
		bidx = SB_INC(bidx);
	}

	/* The range is inclusive */
	sum = NONCE_SUM(tp->scoreboard[bidx].nonce, sum);

	if (sum != nonce_sum) {
		tfc_stats.badnonce++;
		error = TFC_BADNONCE;
		goto out;
	}

	/* Match */
	/* printf("Nonce sum matched for seq %d\n", eseq); */
 out:
	if (error == 0) {
		int acked = SB_INC(eidx);
		tick_t rtto = TICK_DIFF(ts, tp->scoreboard[bidx].ts);
		tp->last_ack_ts = tp->scoreboard[eidx].ts;
		tp->last_ack_seq = eseq;
		acked = SB_DIFF(acked, tp->sb_first);
		update_rtt(rtto, ts, tp);
		tp->nonce_sum = nonce_sum;
		tp->nonce_sumseq = eseq;
		tp->ppl[tp->ppl_open] += acked; 
		tp->sb_first = SB_INC(eidx);
		tp->ack_cur += acked;
		tfc_stats.acktotal += acked;
		/* tp->last_acked = eseq; */
	}
	return error;
}

/*
 * Maybe this should more fancy
 */
static void
update_rtt(tick_t new_rtt, tick_t ts, struct tfc_cb *tp)
{
	
	if (tp->rtt_ignore <= 0 && tp->rtt) {
		tp->rtt = tp->rtt +
			((new_rtt >> RTT_DECAY_SHIFT) - (tp->rtt >> RTT_DECAY_SHIFT));

	} else if (tp->rtt_ignore == 0) {
		/* Initialize rtt */
		tp->rtt = new_rtt;
		printf ("Initial rtt = %lld microseconds\n", TICK2USEC(tp->rtt));
	} else
		tp->rtt_ignore--;

#ifdef AJW_MODIFIED
	if (tp->rtt <= 0)
		tp->rtt = 1;
#endif

	if (tp->rtt <= 0)
	  tp->rtt = new_rtt;

	tp->rto = tp->rtt << 2;
       
	
	/*
	 * Modified by AJW: we now enter slow start as soon as 
	 * a packet is sent
	 */

	/* Enter slow start as soon as we have some estimate of rtt */
	/*
	if (tp->state <= TFCS_INITRTT && tp->rtt) {
		enter_slowstart(tp, ts);
	}
	*/
}

#ifdef AJW_NOT_USED
static void
enter_slowstart(struct tfc_cb *tp, tick_t ts)
{
	tp->state = TFCS_SLOWSTART;
	tp->ack_prev = TFC_OPEN_SS;
	tp->sent_cur = 0;
	tp->ack_cur = 0;
	tp->rate_act = 0;
	tp->sent_prev = 0;
	tp->prev_ts = ts;
	printf("Entering SLOW START: ack_cur = %d, rtt = %lld microseconds\n", tp->ack_cur, TICK2USEC(tp->rtt));
}
#endif /* AJW_NOT_USED */



/*
 * Lookup an element with seqno == seqno
 * Return -1 if not found.
 */
int
sb_lookup(seq_t seq_no, struct tfc_cb *tp)
{
	int i;

	if (tp->sb_lastplus1 == tp->sb_first)
		/* Empty */
		return -1;

	/*
	 * Linear search. Maybe evaluate binary search later
	 */
	for (i = tp->sb_first; i != tp->sb_lastplus1; i = SB_INC(i)) {
		if (tp->scoreboard[i].seq_no == seq_no)
			return i;
		else if (tp->scoreboard[i].seq_no > seq_no)
			break;
	}
	return -1;
}

int
sb_lookup_leq(seq_t seq_no, struct tfc_cb *tp)
{
	int i = tp->sb_first;

	if (tp->sb_lastplus1 == tp->sb_first)
		/* Empty */
		return -1;

	/* only return empty if no scoreboard entry <= seq_no */
	if (tp->scoreboard[tp->sb_first].seq_no > seq_no)
	  return -1;

	/*
	 * Linear search. Maybe evaluate binary search later
	 */
	for (i = SB_INC(i); i != tp->sb_lastplus1; i = SB_INC(i)) {

	  if (tp->scoreboard[i].seq_no > seq_no)
	    return (SB_DEC(i));
	}
	  
	/* return the last element in the scoreboard */
	return SB_DEC(i);
}



void
sb_cleanup(struct tfc_cb *tp)
{
	tp->sb_first = tp->sb_lastplus1;
	return;
}

void
sb_drain(struct tfc_cb *tp, tick_t ts)
{
	return;
}

int
register_loss_leq(seq_t seq_no, struct tfc_cb *tp, int resend)
{
	int sb_idx;
	int new_le_idx;
	int loss_ival;
	int nle = 0;
	int i;


	/*
	 * Look for seq_no in the scoreboard.
	 */
	sb_idx = sb_lookup_leq(seq_no, tp);

	if (sb_idx < 0) {
		printf("Looking for seqno %d but have %d, %d (%d)\n", seq_no, tp->scoreboard[tp->sb_first].seq_no, tp->scoreboard[SB_DEC(tp->sb_lastplus1)].seq_no, SB_DIFF(tp->sb_lastplus1, tp->sb_first));

		return EINVAL;
	}

	/* See if this seq_no is covered by the previous loss-event? -- yes, ignore */
	if (TFC_SEQ_LEQ(seq_no, tp->prev_le_seq) ||
	    TICK_LEQ(tp->scoreboard[sb_idx].ts, tp->prev_le_ts)) {
		return 0;
	}

	/* prev_le_ts is the tick of the first lost packet */
	/* prev_le_seq is the sequence number of the first lost/last acked packet */
	
	/*
	 * TFRC code counts all packets from the last loss-event to the last
	 * received as part of the same loss-event.
	 * It seems right because conventional TCP considers all that as a loss
	 * event and cuts its rate at least by half in the next rtt.
	 */
	if (TFC_SEQ_GEQ(tp->last_ack_seq, seq_no)) {
		/* Last acked is greater than first lost packet */
		/* Loss interval starts there */
		tp->prev_le_seq = tp->last_ack_seq;
		new_le_idx = sb_lookup_leq(tp->last_ack_seq, tp);
	} else {
		/* This loss opens up a new loss interval */
		tp->prev_le_seq = seq_no;
		new_le_idx = sb_idx;
	}

	if (tp->state == TFCS_SLOWSTART) {
		/* First loss takes a connection out of slow-start */
		long long lf;
		long rtt = TICK2USEC(tp->rtt);
		long rto;
		
		/* our equation does not handle large rtts nicely,
		   so send as small as possible, usec or msec
		*/
		if (rtt > 1000) {
			rtt = TICK2MSEC(tp->rtt);
			rto = TICK2MSEC(tp->rto);
		} else {
			rto = TICK2USEC(tp->rto);
		}

		tp->state = TFCS_STEADYSTATE;
		printf("Entering steadystate at seq %d, ack_prev = %d, rtt = %lld\n", seq_no, tp->ack_prev, (long long)tp->rtt);
		lf = fake_rate(tp->rate_eq, rtt, rto);
		tp->ppl_open = 0;
		loss_ival = (int)lf;
		printf("fake loss interval = %d ok\n", loss_ival);
		tp->sent_cur = 0;
	} else {

		/*
		  from sb_start till the last_loss, everything is part of the current
		  open loss interval
		*/
		loss_ival = SB_DIFF(new_le_idx, tp->prev_le_idx);
	}

	/* New first loss seq and timestamp */
	tp->prev_le_idx = new_le_idx;

	/* Check if more than one loss events have happened
	 * Assume uniform loss rate in the previous rtts
	 */
	if (TICK_GT(tp->scoreboard[sb_idx].ts, tp->last_ack_ts)) {
		nle = TICK_DIFF(tp->scoreboard[sb_idx].ts, tp->last_ack_ts);


		/* XXX AJW: avoid divide by zero errors.  is this right? */
		if (tp->rtt == 0) {
		  printf("warning: zero rtt\n");
		  nle = 1;
		}
		else nle = nle/tp->rtt;
	}
	if (!nle)
		nle = 1;
	else
		loss_ival = loss_ival/nle;
	
	tp->prev_le_ts = tp->scoreboard[new_le_idx].ts + tp->rtt;
	
	/* ppl update */
	for (i = 0; i < nle; i++) {
		tp->ppl[tp->ppl_open] = loss_ival;
		tp->ppl_open = PPL_INC(tp->ppl_open);
		tp->ppl[tp->ppl_open] = 0;
	}
	
	/*
	 * Assume everything upto new_le_idx has been acked.
	 * New sb_start. sb_start should change in response to nacks and acks
	 */
	if ( !resend) {
		tp->sb_first = SB_INC(new_le_idx);		
		tp->nonce_sumseq = tp->prev_le_seq; /* make nonce_sum invalid */
		tp->nonce_sumsynched = 0;
		tp->last_ack_seq = tp->prev_le_seq;
		tp->last_ack_ts = tp->prev_le_ts;
	}
	
	tp->checknow = 0; /* Check the sender after this loss */
	/* printf("Loss: ppl_open %d, ppl %d, sumseq %d\n",  tp->ppl_open, loss_ival,
	   tp->nonce_sumseq);
	*/
	return 0;
}


/*
 * Register a loss. Update the ppl and scoreboard.
 */
/* Update the last loss interval */
 
/*
 * We may need size of the packet. Our loss event rate should be the ratio of bytes
 * sent to bytes acked. We cannot do packets like TFRC because packets are not fixed sized.
 */
/*
 * XXX: Maybe a loss event should be split into more than one if it encompasses
 * more than one rtt periods. For example, if a loss is reported after three rtts then it
 * may not increase the size of the loss-interval.
 */
int
register_loss(seq_t seq_no, struct tfc_cb *tp, int resend)
{
	int sb_idx;
	int new_le_idx;
	int loss_ival;
	int nle = 0;
	int i;
	/*
	 * Look for seq_no in the scoreboard.
	 */
	sb_idx = sb_lookup(seq_no, tp);
	if (sb_idx < 0) {
#ifdef SB_DEBUG
		printf("Looking for seqno %d but have %d, %d (%d)\n", seq_no, tp->scoreboard[tp->sb_first].seq_no, tp->scoreboard[SB_DEC(tp->sb_lastplus1)].seq_no, SB_DIFF(tp->sb_lastplus1, tp->sb_first));
#endif
		return EINVAL;
	}

	/* See if this seq_no is covered by the previous loss-event? -- yes, ignore */
	if (TFC_SEQ_LEQ(seq_no, tp->prev_le_seq) ||
	    TICK_LEQ(tp->scoreboard[sb_idx].ts, tp->prev_le_ts)) {
		return 0;
	}

	/* prev_le_ts is the tick of the first lost packet */
	/* prev_le_seq is the sequence number of the first lost/last acked packet */
	
	/*
	 * TFRC code counts all packets from the last loss-event to the last
	 * received as part of the same loss-event.
	 * It seems right because conventional TCP considers all that as a loss
	 * event and cuts its rate at least by half in the next rtt.
	 */
	if (TFC_SEQ_GEQ(tp->last_ack_seq, seq_no)) {
		/* Last acked is greater than first lost packet */
		/* Loss interval starts there */
		tp->prev_le_seq = tp->last_ack_seq;
		new_le_idx = sb_lookup(tp->last_ack_seq, tp);
	} else {
		/* This loss opens up a new loss interval */
		tp->prev_le_seq = seq_no;
		new_le_idx = sb_idx;
	}

	if (tp->state == TFCS_SLOWSTART) {
		/* First loss takes a connection out of slow-start */
		long long lf;
		long rtt = TICK2USEC(tp->rtt);
		long rto;
		
		/* our equation does not handle large rtts nicely,
		   so send as small as possible, usec or msec
		*/
		if (rtt > 1000) {
			rtt = TICK2MSEC(tp->rtt);
			rto = TICK2MSEC(tp->rto);
		} else {
			rto = TICK2USEC(tp->rto);
		}

		tp->state = TFCS_STEADYSTATE;
		printf("Entering steadystate at seq %d, ack_prev = %d...\n", seq_no, tp->ack_prev);
		lf = fake_rate(tp->ack_prev, rtt, rto);
		tp->ppl_open = 0;
		loss_ival = (int)lf;
		printf("ival = %d ok\n", loss_ival);
		tp->sent_cur = 0;
	} else {

		/*
		  from sb_start till the last_loss, everything is part of the current
		  open loss interval
		*/
		loss_ival = SB_DIFF(new_le_idx, tp->prev_le_idx);
	}

	/* New first loss seq and timestamp */
	tp->prev_le_idx = new_le_idx;

	/* Check if more than one loss events have happened
	 * Assume uniform loss rate in the previous rtts
	 */
	if (TICK_GT(tp->scoreboard[sb_idx].ts, tp->last_ack_ts)) {
		nle = TICK_DIFF(tp->scoreboard[sb_idx].ts, tp->last_ack_ts);

		/* XXX AJW: avoid divide by zero errors.  is this right? */
		if (tp->rtt == 0)
		  nle = 1;
		else nle = nle/tp->rtt;

	}
	if (!nle)
		nle = 1;
	else
		loss_ival = loss_ival/nle;
	
	tp->prev_le_ts = tp->scoreboard[new_le_idx].ts + tp->rtt;
	
	/* ppl update */
	for (i = 0; i < nle; i++) {
		tp->ppl[tp->ppl_open] = loss_ival;
		tp->ppl_open = PPL_INC(tp->ppl_open);
		tp->ppl[tp->ppl_open] = 0;
	}
	
	/*
	 * Assume everything upto new_le_idx has been acked.
	 * New sb_start. sb_start should change in response to nacks and acks
	 */
	if ( !resend) {
		tp->sb_first = SB_INC(new_le_idx);		
		tp->nonce_sumseq = tp->prev_le_seq; /* make nonce_sum invalid */
		tp->nonce_sumsynched = 0;
		tp->last_ack_seq = tp->prev_le_seq;
		tp->last_ack_ts = tp->prev_le_ts;
	}
	
	tp->checknow = 0; /* Check the sender after this loss */
	/* printf("Loss: ppl_open %d, ppl %d, sumseq %d\n",  tp->ppl_open, loss_ival,
	   tp->nonce_sumseq);
	*/
	return 0;
}

/***** Functions that work with byte ranges *****/

/*
 * There are a number of invariants that must be satisfied by the seq number range:
 *   for example, eseq >= bseq and
                  eseq > eseq of the newest record in the scoreboard.
*/

int
sb_insert_range(struct tfc_cb* tp, seq_t bseq, seq_t eseq, u_char nonce, 
		tick_t ts)
{
	/* Enter this packet in the scoreboard */
	int nlp1 = SB_INC(tp->sb_lastplus1);
	
	if (nlp1 == tp->sb_first) {
		/* Overflow */
		
		/* printf ("Scoreboard overflow for seq %d\n", n); */
		tfc_stats.sb_overflows++;

		/* XXX: Maybe cope with the overflow */
		return TFC_SBOF;
	}
#ifdef SB_TRACE
	printf("SC in %d, %d at index %d\n", bseq, eseq, tp->sb_lastplus1);
#endif
	tp->scoreboard[tp->sb_lastplus1].bseq = bseq;
	tp->scoreboard[tp->sb_lastplus1].eseq = eseq;
	tp->scoreboard[tp->sb_lastplus1].nonce = nonce;
	tp->scoreboard[tp->sb_lastplus1].ts = ts;
	tp->scoreboard[tp->sb_lastplus1].state = SBSTATE_SENT;
	tp->sb_lastplus1 = nlp1;
	return 0;
}

/* match sum, the range is inclusive of sb_first and eidx */
int
sb_matchsum_range(struct tfc_cb *tp, seq_t bseq, seq_t eseq, u_char nonce_sum, tick_t ts)
{
	int error = 0;
	int bidx = -1;
	int eidx = -1;
	int isbeg = 0, isend = 0;
	int npackets = 0;
	u_char sum = tp->nonce_sum;

#ifdef SB_TRACE
	printf("Matching sum for range: %d, %d\n", bseq, eseq);
#endif
	/*
	 * If the seq num has already been covered by a previous ack
	 * (also indicated by an empty scoreboard)
	 * then don't do anything.
	 */
	if (tp->sb_first == tp->sb_lastplus1 ||
	    SEQ_LT(eseq, tp->scoreboard[tp->sb_first].bseq))
		return 0;
	
	/*
	 * Look up the beginning sequence number.
	 */
	npackets = sb_lookup_range(tp, bseq, eseq, &bidx, &isbeg, &eidx, &isend);
	if (npackets < 0) {
		error = TFC_EINVALSEQ;
		goto out;
	}

	if (bidx != tp->sb_first) {
		/* The sum can only be matched in-order */
		error = TFC_BADNONCE;
		goto out;
	}

	/* Calculate the sum */
	while(bidx != eidx) {
		sum = NONCE_SUM(tp->scoreboard[bidx].nonce, sum);
		tp->scoreboard[bidx].state = SBSTATE_ACKED;
		bidx = SB_INC(bidx);
	}
	if (isend)
		/* The range is inclusive of eidx */
		sum = NONCE_SUM(tp->scoreboard[bidx].nonce, sum);

#ifdef SB_TRACE
	printf("\t%d packets\n", npackets);
#endif
	/*
	 * XXX: A transport should not be allowed to ack more than a
	 * single rtt's worth of data in one shot. Or else multiple loss events
	 * may get counted as a single one. 
	 */ 
	if (tp->nonce_sumsynched == 0) {
		tp->nonce_sumsynched = 1;
	} else if (sum != nonce_sum) {
		tfc_stats.badnonce++;
#ifdef SB_TRACE
		printf("\tMismatch\n");
#endif
		error = TFC_BADNONCE;
		goto out;
	}	
	/* Match */
 out:
	if (error == TFC_EINVALSEQ) {
#ifdef SB_TRACE
		printf("\tInvalid sequence number\n");
#endif
		return error;
	}
	
	if (error == 0) {
		/* observed rtt */
		tick_t rtto = TICK_DIFF(ts, tp->scoreboard[bidx].ts);
		update_rtt(rtto, ts, tp);
		tp->last_ack_ts = tp->scoreboard[eidx].ts;
		tp->last_ack_seq = eseq;
		tp->nonce_sum = nonce_sum;
		tp->nonce_sumseq = eseq;
		tp->ack_cur += npackets;

		if (tp->state == TFCS_SLOWSTART)
		  tp->slow_start_credits += (npackets * 2);

		tfc_stats.acktotal += npackets;
		/* Exactly where should I do these? */
		sb_advance(tp, eseq, eidx, isend);
		tp->ppl[tp->ppl_open] += npackets;
	}
#ifdef SB_TRACE
	printf("\tError = %d...Done\n", error);
#endif
	return error;
}

int
sb_lookup_range(struct tfc_cb *tp, seq_t bseq, seq_t eseq, int *bidx, int *isbeg,
		int *eidx, int *isend)
{
	int i;
	int count = 0;
	int bfound = 0; 

	if (tp->sb_lastplus1 == tp->sb_first)
		/* Empty */
		return -1;

	i = tp->sb_first;
	if (SEQ_LT(bseq, tp->scoreboard[i].bseq) || SEQ_LT(eseq, tp->scoreboard[i].bseq))
		return -1;
	/*
	 * Linear search. 
	 */
	for (; i != tp->sb_lastplus1; i = SB_INC(i)) {

		if (!bfound) {
			if (tp->scoreboard[i].bseq == bseq) {
				*isbeg = 1;
				*bidx = i;
				bfound = 1;
			} else if (SEQ_LT(bseq, tp->scoreboard[i].bseq) ||
				   SEQ_GT(tp->scoreboard[i].eseq, bseq) ||
				   tp->scoreboard[i].eseq == bseq) {
				*isbeg = 0;
				*bidx = i;
				bfound = 1;
			}
		}
		
		if (bfound)
			count++;
		
		if (tp->scoreboard[i].eseq == eseq) {
			*isend = 1;
			break;
		}
		else if (SEQ_GT(tp->scoreboard[i].eseq, eseq)) {
			*isend = 0;
			break;
		}
	}
	/* bseq or eseq is not present in the scoreboard */
	if (!bfound || i == tp->sb_lastplus1)
		return -1;

	*eidx = i;
	return count;
}

int
sb_advance(struct tfc_cb *tp, seq_t eseq, int idx, int isboundary)
{
	if (!isboundary) {
		tp->sb_first = idx;
		tp->scoreboard[idx].bseq = eseq + 1;
	} else {
		tp->sb_first = SB_INC(idx);
	}
	return idx;
}

int
register_loss_range(struct tfc_cb *tp, seq_t bseq, seq_t eseq, int resend)
{
	int bidx, eidx, eidxplus1 = 0;
	int i;
	int isbeg = 0, isend = 0, npackets;

#ifdef SB_TRACE
	printf("register loss %d, %d\n", bseq, eseq);
#endif
	/*
	 * Get the range of lost sequences in the scorebord
	 */
	npackets = sb_lookup_range(tp, bseq, eseq, &bidx, &isbeg, &eidx, &isend);
	if (npackets < 0) {
#ifdef SB_TRACE
		/*		printf("Looking for seqno %d, %d but have %d, %d (%d)\n", bseq, eseq, tp->scoreboard[tp->sb_first].bseq, tp->scoreboard[SB_DEC(tp->sb_lastplus1)].bseq, SB_DIFF(tp->sb_lastplus1, tp->sb_first)); */
		printf ("unable to find seq %d, %d in sb\n", bseq, eseq);
#endif
		return EINVAL;
	}
	
	/* Mark all packets in the range */
	i = bidx;
	while (i != eidx) {
		tp->scoreboard[i].state = SBSTATE_LOST;
		i = SB_INC(i);
	}
	tp->scoreboard[i].state = SBSTATE_LOST;

	/*
	 * If we are not resending then extend the range
	 * as much as possible by including further lost packets
	 */ 
	eidxplus1 = SB_INC(eidx);
	if (!resend) {
		/* out of order packets are lost */
		if (bidx != tp->sb_first)
			return 0;
		
		while (eidxplus1 != tp->sb_lastplus1 &&
		       (tp->scoreboard[eidxplus1].state == SBSTATE_ACKED ||
			tp->scoreboard[eidxplus1].state == SBSTATE_LOST)) {
			eidx = eidxplus1;
			eidxplus1 = SB_INC(eidxplus1);
		}
	}

	/* prev_le_ts is the tick of the first lost packet + rtt */
	/* prev_le_seq is the sequence number of the first lost packet */

	/* update ppl */
	if (tp->state < TFCS_SLOWSTART) {
		/* we do not even have an estimate of rtt */
		tp->prev_le_ts = tp->scoreboard[eidx].ts;
		tp->prev_le_seq = tp->scoreboard[eidx].eseq;
		goto out;
	} else if (tp->state == TFCS_SLOWSTART) {
		/* First loss takes a connection out of slow-start */
		long long lf;
		long rtt = TICK2USEC(tp->rtt);
		long rto;
		int rate;
		
		/* our equation does not handle large rtts nicely,
		   so send as small as possible, usec or msec
		*/
		if (rtt > 1000) {
			rtt = TICK2MSEC(tp->rtt);
			rto = TICK2MSEC(tp->rto);
		} else {
			rto = TICK2USEC(tp->rto);
		}

		tp->state = TFCS_STEADYSTATE;
		printf("Entering steadystate at seq %d, prev rate = %d, rtt = %lld\n", eseq,
		       tp->rate_eq, (long long)tp->rtt);
		rate = max((int)tp->rate_eq/2, tp->ack_prev);
		lf = fake_rate(rate, rtt, rto);
		tp->ppl[0] = (int) lf;
		printf(" fake loss interval = %d\n", tp->ppl[0]);
		tp->ppl_open = PPL_INC(0);
		tp->sent_cur = 0;
		tp->prev_le_ts = TICK_ADD(tp->scoreboard[eidx].ts, tp->rtt);
		tp->prev_le_seq = tp->scoreboard[eidx].eseq;
	} else {		
		/* Start from the bidx and go up to one rtt, all these packets are part of
		   one loss event. Similarly, compute loss events in this entire range.
		*/
		while (bidx != eidxplus1) {
			if(TICK_GT(tp->scoreboard[bidx].ts, tp->prev_le_ts) &&
			   tp->scoreboard[bidx].state == SBSTATE_LOST) {
				tp->prev_le_ts = TICK_ADD(tp->scoreboard[bidx].ts, tp->rtt);
				tp->prev_le_seq = tp->scoreboard[bidx].eseq;
				tp->ppl_open = PPL_INC(tp->ppl_open);
				tp->ppl[tp->ppl_open] = 0;
			} 
			bidx = SB_INC(bidx);
			tp->ppl[tp->ppl_open]++;
		} 
	}

 out:
	/*
	 * if we are not resending, then adjust scoreboard to ignore the
	 * lost packets
	 */
	if ( !resend) {
		tp->sb_first = eidxplus1;		
		tp->nonce_sumseq = tp->prev_le_seq; /* make nonce_sum invalid */
		tp->nonce_sumsynched = 0;
		tp->last_ack_seq = tp->prev_le_seq;
		tp->last_ack_ts = tp->prev_le_ts;
	}
	
	/* Check the sender after a loss */
	tp->checknow = 0;
#ifdef SB_TRACE
	printf("\tLoss exit: ppl_open %d, ppl %d, sumseq %d\n",  tp->ppl_open,
	       tp->ppl[PPL_DEC(tp->ppl_open)], tp->nonce_sumseq);
#endif
	return 0;
}

int
sb_update_range(struct tfc_cb *tp, seq_t bseq, seq_t eseq, tick_t ts, u_char *nonce)
{
	int bidx, isb, eidx, ise, idx;
	int npackets = sb_lookup_range(tp, bseq, eseq, &bidx, &isb, &eidx, &ise);
	u_char sum = 0;
	
	if (npackets < 0)
		return EINVAL;
	
	if (npackets == 1) {
		/* a single packet */
#ifdef SB_TRACE		
		if (!isb || !ise) {
			printf("%s:TSNHH:\n"
			       "A fragment of a previously sent packet is being resent\n", __FILE__);
		}
#endif
		tp->scoreboard[bidx].ts = ts;
		tp->scoreboard[bidx].state = SBSTATE_RESENT;
		sum = NONCE_SUM(tp->scoreboard[bidx].nonce, sum);
	} else {
		idx = bidx;
		while (idx != eidx) {
			tp->scoreboard[idx].ts = ts;
			tp->scoreboard[idx].state = SBSTATE_RESENT;
			sum = NONCE_SUM(tp->scoreboard[idx].nonce, sum);
			idx = SB_INC(idx);
		}
		tp->scoreboard[eidx].ts = ts;
		tp->scoreboard[eidx].state = SBSTATE_RESENT;
		if (ise)
			sum = NONCE_SUM(tp->scoreboard[eidx].nonce, sum);
	}
	*nonce = sum;
	return 0;
}
