/*
 * tfc.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/proc.h>
#include <sys/jail.h>
#include <sys/conf.h>

#include <vm/vm_zone.h> /* memory for tfc_cb and the likes */

#include "opt_xtcp.h"

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_pcb.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>

#include <netinet/xtcp/tfc.h>
#include <netinet/xtcp/tfci.h>
#include <netinet/xtcp/tfi.h>
#include <netinet/xtcp/tfc_stats.h>
#include <netinet/xtcp/tfc_error.h>
#include <netinet/xtcp/lossfraction.h>
#include <netinet/xtcp/scoreboard.h>
#include <netinet/xtcp/tfc_cb.h>
#include <netinet/xtcp/tfi_pcb.h>

#include <netinet/xtcp/xldev.h>

struct tfc_stats tfc_stats;

int tfc_2msl;
int tfc_nofeedinit = 1;

/* Globals */
extern struct domain inetdomain;

/*
 * For now let there be only one domain. Maybe later we'll have a
 * dynamically expandable list of domains.
 */
static int tfc_usr_attach __P((struct socket *so, int proto, struct proc *p));
static int tfc_usr_detach __P((struct socket *so));
static int tfc_usr_abort __P((struct socket *so));
static int tfc_usr_accept __P((struct socket *so, struct sockaddr **nam));
static int tfc_usr_bind __P((struct socket *so, struct sockaddr *nam, struct proc *p));
static int tfc_usr_connect __P((struct socket *so, struct sockaddr *nam, struct proc *p));
static int tfc_usr_disconnect __P((struct socket *so));
static int tfc_usr_listen __P((struct socket *so, struct proc *p));
static int tfc_usr_rcvd __P((struct socket *so, int flags));
static int tfc_usr_rcvoob __P((struct socket *so, struct mbuf *m, int flags));
static int tfc_usr_send __P((struct socket *so, int flags, struct mbuf *m, 
			 struct sockaddr *nam, struct mbuf *control, struct proc *p));
static int tfc_usr_shutdown __P((struct socket *so));
	
struct pr_usrreqs tfc_usrreqs = {
	tfc_usr_abort, tfc_usr_accept, tfc_usr_attach, tfc_usr_bind,
	tfc_usr_connect, pru_connect2_notsupp, in_control, tfc_usr_detach,
	tfc_usr_disconnect, tfc_usr_listen, in_setpeeraddr, tfc_usr_rcvd,
	tfc_usr_rcvoob, tfc_usr_send, pru_sense_null, tfc_usr_shutdown,
	in_setsockaddr, sosend, soreceive, sopoll
};

/* Internal functions */
static int tfc_detach __P((register struct tfc_cb *tp));
void tfc_timer_2msl __P((void *xtp));
int tfc_drop __P((struct tfc_cb *tp, int errno));
static int tfc_attach __P((struct socket *so, struct inpcbinfo *xtcbinfo, struct proc *p));
void tfc_close __P((struct tfc_cb *tp));
void tfc_notify __P((struct inpcb *inp, int error));

#define PARANOIA /* We may need this because Cyclone exception handling code is
		    non-reentrant.
		 */
#define VERBOSE_DEBUG

void
tfc_init(void)
{
	int hashsize = TFCCB_HASHTABLE_SIZE; /* safe default */
	struct inpcbinfo *xtcbinfo;
	struct	vm_zone *xzone;
	int i = 0;
	
#ifdef TFC_LATER	
	TUNABLE_INT_FETCH("net.inet.tcp.tcbhashsize", &hashsize);
	if (!powerof2(hashsize)) {
		printf("WARNING: TCB hash size not a power of 2\n");
		hashsize = 512; /* safe default */
	}
#endif
	tfc_2msl = 120 * hz;
	xzone = zinit("tfc_cb", sizeof(struct inp_xtcb), maxsockets,
				 ZONE_INTERRUPT, 0);

	/* Do the following for all domains */
	for (i = 0; i < TFC_NUM_DOMAINS; i++) {
		LIST_INIT(&tfc_domains[i]->xtcb);
		xtcbinfo = &tfc_domains[i]->xtcbinfo;
		xtcbinfo->listhead = &tfc_domains[i]->xtcb;
		
		/* Hash tables to store the port->tcb mapping */
		xtcbinfo->hashbase = hashinit(hashsize, M_PCB, &xtcbinfo->hashmask);
		xtcbinfo->porthashbase = hashinit(hashsize, M_PCB,
						  &xtcbinfo->porthashmask);
		xtcbinfo->ipi_zone = xzone;
		tfc_domains[i]->hashbase =
			hashinit(tfc_domains[i]->hashsize, M_PCB, &tfc_domains[i]->hashmask);
		tfc_domains[i]->check_hdr = tfc_check_tphdr;
	}
	/* Done! */
	printf("TFC: TFC successfully initialized!\n");
}

int
tfc_check_tphdr(struct mbuf **pm, struct tfc_tp *tp, int off0)
{
	struct tfc_tp *th;
	int iphlen = off0;

	if ((*pm)->m_len < iphlen + sizeof (struct tfc_tp)) {
		if ((*pm = m_pullup(*pm, iphlen + sizeof(struct tfc_tp))) == 0) {
			tfc_stats.xtp_rcvshort++;
			printf("Unable to pull up generic XTPIP header\n"); 
			return EINVAL;
		}
	}	
	th = (struct tfc_tp *)((caddr_t)mtod(*pm, caddr_t) + iphlen);
	tp->th_sport = th->th_sport;
	tp->th_dport = th->th_dport;
	return 0;
}

void
tfc_input(struct mbuf *buf, int off0, int proto)
	/* off0: offset of data in IP datagram buf */
{
	/*
	 * find the protocol
	 * and call its input
	 */
	struct ip *ip;
	struct tfi_protosw *psw;
	struct inpcb *inp;
	struct tfc_cb *tp;
	struct tfc_protodomain *dm;
	struct tfc_tp th;
	int error = 0;
#ifdef PARANOIA
	int s = splnet();
#endif
	
	tfc_stats.recvtotal++;
	dm = protolookup_domain(proto);
	if (dm == NULL) {
		/* This should not have happened */
		tfc_stats.nodmdrops++;
		error = TFC_ENOSUPPORT;
		goto error_out;
	}

	/*
	 * Get IP and transport header together in first mbuf.
	 * Note: IP leaves IP header in first mbuf.
	 */
	if (off0 > sizeof (struct ip)) {
		ip_stripoptions(buf, (struct mbuf *)0);
		off0 = sizeof(struct ip);
	}

	error = dm->check_hdr(&buf, &th, off0);
	if (error != 0) {
		goto error_out;
	}
	
	ip = mtod(buf, struct ip *);
	/*
	 * Checksum should be computed by the transport protocol code.
	 */
	inp = in_pcblookup_hash(&dm->xtcbinfo, ip->ip_src, th.th_sport,
	    ip->ip_dst, th.th_dport, 1, buf->m_pkthdr.rcvif);

	if (inp == NULL) {
		/*
		 * We do not have a protocol connected to this port.
		 *  Silently drop the packet.
		 */
		tfc_stats.noport_drops++;
		error = TFC_NOPROTO;
		goto error_out;
	}
	tp = intotfccb(inp);
	psw = tp->psw;
	if (NULL == psw) {
		/* Free the buffer and bailout */
		tfc_stats.noproto_drops++;
		error = TFC_NOPROTO;
		goto error_out;
	}

	error = tfi_upcall(psw->pr_input, buf, off0, proto, inp->inp_socket, tp->cyc_psw);
	if (error)
		goto error_out;
	else
		goto out;	
	/*
	 * The TFC layer should find out whether the input function added any data to the
	 * socket. If yes, it should invoke the function that was needed to be invoked.
	 */
	/*
	 * User transport should not be allowed to wakeup a process
	 * Is this a security concern?
	 * Maybe not, it can only wakeup the process it is connected to.
	 */
 error_out:
	printf ("Receive error: %d\n", error);
	tfc_stats.input_drops++;
	if (buf) {
		m_freem(buf);
	}
 out:
#ifdef PARANOIA
	splx(s);
#endif
	/* printf("recv done\n"); */
	return;
}

void
tfc_notify(struct inpcb *inp, int error)
{
	struct tfc_cb *tp = intotfccb(inp);
#ifdef PARANOIA
	int s = splnet();
#endif
	if (tp == NULL || tp->psw == NULL)
#ifdef PARANOIA
	{
		splx(s);
#endif
		return;
#ifdef PARANOIA
	}
#endif
	tfi_upcallv(tp->psw->pr_ctlinput, inp->inp_socket, error, tp->cyc_psw);
#ifdef PARANOIA
	splx(s);
#endif
	return;
}

void
tfc_ctlinput(cmd, sa, vip)
	int cmd;
	struct sockaddr *sa;
	void *vip;
{
	struct in_addr faddr;
	int s, i;
		
	faddr = ((struct sockaddr_in *)sa)->sin_addr;
	if (sa->sa_family != AF_INET || faddr.s_addr == INADDR_ANY)
		return;

	s = splnet();

	/* for each domain based in INET, do the following */
	for (i = 0; i < TFC_NUM_DOMAINS; i++) {
		in_pcbnotifyall(&tfc_domains[i]->xtcb, faddr, inetctlerrmap[cmd],
				tfc_notify);
	}
	splx(s);
	return;
}

/*
 * XXX According to a comment in tcp_usrreq.c this code need not run
 * at splnet anymore. But the code is still there.
 */
int
tfc_ctloutput(so, sopt)
	struct socket *so;
	struct sockopt *sopt;
{
	int error, s, i;
		
	error = 0;
	s = splnet();		/* XXX */

	error = 1;
	for (i = 0; i < TFC_NUM_DOMAINS; i++) {
		if (sopt->sopt_level == tfc_domains[i]->proto) {
			error = 0;
			break;
		}
	}

	/* Not for any of our protocols */
	if (error) {
		error = ip_ctloutput(so, sopt);
		goto out;
	}
	
	switch (sopt->sopt_dir) {
		/*
		 * TFC layer may handle some options itself before passing
		 * anything down to user transport
		 */
	case SOPT_SET:
		if ((sopt->sopt_name == TFC_ATTACH_PROTOCOL) ||
		    (sopt->sopt_name == TFC_ATTACH_PROTOCOL_EXCL)) {
			struct inpcb *inp;
			u_long hash;
			struct tfc_cb *tp;
			struct tfc_proto *xp;
			
			error = sooptcopyin(sopt, &hash, sizeof hash,
					    sizeof hash);			
			if (0 != error)
				goto out;

			printf("Tfc option attach: %u received\n", (u_int)hash);
			/* Check if a protocol is already bound to this socket */
			inp = sotoinpcb(so);
			if (inp == NULL) {
				/* this should not have happened */
				error = TFC_EARGS;
				goto out;
			}
			tp = intotfccb(inp);

			if (tp == NULL) {
				/* this should not have happened */
				error = TFC_EARGS;
				goto out;
			}
			
			if (tp->xp != NULL || tp->psw != NULL) {
				error = TFC_EXISTS;
				goto out;
			}
			/* Check if the "hash" protocol is loaded */
			if ((xp = domain_findproto(tp->dm, hash)) == NULL) {
				struct in_addr ia;
				
				ia.s_addr = htonl(0x7F000001); /* XXX */
				xldev_req_append_in(xld_sdev, &ia, hash);

				/* Should we load the default protocol?? */
				if (sopt->sopt_name == TFC_ATTACH_PROTOCOL_EXCL) {
					error = TFC_ENOSUPPORT;
					goto out;
				} else {
					/* load the default protocol */
					if (tp->dm->default_protop == NULL) {
						error = TFC_ENOSUPPORT;
						goto out;
					} else
						xp = tp->dm->default_protop;
				}
				
			}
#ifdef TFC_NEGOTIATE
			if (tp->dm->default_protop != xp && tp->dm->default_protop) {
				tp->requested_proto = xp;
				xp = tp->dm->default_protop;
			} else if (tp->dm->default_protop) /* only for testing */
				tp->requested_proto = xp;
#endif /* TFC_NEGOTIATE */
			error = tfc_soattachproto(so, tp->dm->proto, xp,
						  sopt->sopt_p);
			break;
		}
		/* XXX: Call the transport function to handle this sockopt */
		break;
	case SOPT_GET:
	{
		struct inpcb *inp;
		struct tfc_cb *tp;
		u_long hash;
		
		inp = sotoinpcb(so);
		if (inp == NULL) {
				/* this should not have happened */
			error = TFC_EARGS;
			goto out;
		}
		tp = intotfccb(inp);
			
		if (tp == NULL) {
				/* this should not have happened */
			error = TFC_EARGS;
			goto out;
		}
			
			
		switch (sopt->sopt_name) {
		case TFC_PROTONAME:
			if (tp->xp)
				hash = tp->xp->hash;
			else
				hash = 0;
			break;
		default:
			error = ENOPROTOOPT;
			break;
		}
		if (error == 0)
			error = sooptcopyout(sopt, &hash, sizeof hash);
		break;
	}
	}
 out:
	splx(s);
	return (error);
}

int
tfc_soattachproto(struct socket* so, int proto, struct tfc_proto *xp,
		   struct proc *p) 
{
	int error = 0;
#ifdef PARANOIA
	int s = splnet();
#endif
	struct tfc_cb *tp = sototfccb(so);
	if (tp == NULL)
        {
#ifdef PARANOIA
		splx(s);
#endif
		return EINVAL;
	}
	error = tfi_upcall(xp->psw->pr_usrreqs->pru_attach, so, proto, p, xp->cyc_psw);
#ifdef PARANOIA
	splx(s);
#endif
	
	if (error)
		return error;
	tp->psw = xp->psw;
	tp->cyc_psw = xp->cyc_psw;
	tp->xp = xp;
	tp->state = TFCS_ATTACHED;
	xp->ref_count += 1;
	tfc_stats.protoattached++;
	return 0;
}

void
tfc_fasttimo(void)
{
	/* Maybe check to see if any user transport function is running
	 * for too long...
	 * XXX will slowtimo be called if that happens??
	 */
	tfc_stats.fasttimo++;
	return;
}

void
tfc_slowtimo(void)
{
	/*
	 * May be do some book-keeping here
	 */
	tfc_stats.slowtimo++;
	return;
}

void
tfc_drain(void)
{
	/* Free some mbufs, if possible */
	return;
}

/* This function is called in response to "socket" or "accept" syscalls */
static int
tfc_usr_attach(struct socket *so, int proto, struct proc *p)
{
	/*
	 * Find out the corresponding protocol handler and
	 * let it handle the request.
	 */
	int s = splnet();
	int error = 0;
	struct inpcb *inp;
	struct tfc_cb *tp;
	struct tfc_protodomain *dm;
	
	inp = sotoinpcb(so);
	if (inp != 0) {
		/* Already connected */
		error = TFC_EISCONN;
		goto out;
	}

	/* sonewconn calls pru_attach with proto == 0 */
	if (proto == 0) {
		if (so->so_head)
			dm = sototfccb(so->so_head)->dm;
		else
			dm = NULL;
	} else 
		dm = protolookup_domain(proto);

	if (0 == dm) {
		error = TFC_ENOSUPPORT;
		goto out;
	}

	/* allocate memory for inpcb, tfc_cb */
	error = tfc_attach(so, &dm->xtcbinfo, p);
	if (error)
		goto out;

	
	/* Set the linger time for the socket */
	if ((so->so_options & SO_LINGER) && so->so_linger == 0)
		so->so_linger = TCP_LINGERTIME;

	/* There is no transport attached with this socket as yet. */
	/* In sosetopt there should be a call to initialize the transport
	 * state about this new connection
	 */

	/* Attach the domain to the protocol, e.g., tfc_tcpd */
	tp = sototfccb(so);
	tp->dm = dm;
	tp->state = TFCS_INIT;

 out:
	printf("Attach: %d\n", error);
	splx(s);
	return error;
}

u_long tfc_sendspace = 1024*32;

#ifdef TFC_LATER 
SYSCTL_INT(_net_inet_tfc, TCPCTL_SENDSPACE, sendspace, CTLFLAG_RW, 
    &tcp_sendspace , 0, "Maximum outgoing TCP datagram size");
#endif

u_long tfc_recvspace = 57344;	/* largest multiple of PAGE_SIZE < 64k */

#ifdef TFC_LATER
SYSCTL_INT(_net_inet_tfc, TCPCTL_RECVSPACE, recvspace, CTLFLAG_RW, 
    &tcp_recvspace , 0, "Maximum incoming TCP datagram size");
#endif

static int
tfc_attach(struct socket *so, struct inpcbinfo *xtcbinfo, struct proc *p)
{
	register struct tfc_cb *tp;
	struct inpcb *inp;
	int error;

	/* This should be done by the protocol */
	/*
	if (so->so_snd.sb_hiwat == 0 || so->so_rcv.sb_hiwat == 0) {
		error = soreserve(so, tfc_sendspace, tfc_recvspace);
		if (error)
			return (error);
	}
	*/
	error = in_pcballoc(so, xtcbinfo, p);
	if (error) {
		printf("No pcb\n"); 
		return (error);
	}
	inp = sotoinpcb(so);
	inp->inp_vflag |= INP_IPV4;
	tp = tfc_newtfccb(inp);
	if (tp == 0) {
		int nofd = so->so_state & SS_NOFDREF;	/* XXX */

		so->so_state &= ~SS_NOFDREF;	/* don't free the socket yet */
		printf("No tfccb. detaching inpcb...\n"); 
		in_pcbdetach(inp);
		printf("ok\n"); 
		so->so_state |= nofd;
		return (ENOBUFS);
	}
	/*
	 * Maybe I should add a new socket state that indicates the state
	 * regarding the protocol attachment
	 * Something like, SS_XTPROTOATTAHED.
	 */
	return (0);
}

/*
 * The xtcb is freed after 2msl irrespective of the will of the user protocol
 */
static int
tfc_usr_detach(struct socket *so)
{
	int s = splnet();
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tfc_cb *tp;

	if (inp == 0) {
		splx(s);
		return TFC_EINVAL;
	}
	
	tp = intotfccb(inp);
	if (tp == 0) {
		splx(s);
		return TFC_EINVAL;
	}
	if (tp->state >= TFCS_ATTACHED)
		error = tfc_detach(tp);
	else
		error = tfc_drop(tp, 0);

	splx(s);

	printf("Total packets sent: %d\n\tgot acked %d\n\tgot nacked %d\n\tbad nonce %d\n"
	       "Total packets recvd: %d\n\tdropped %d\n\tacksent %d\n"
	       "SB overflows: %d\n"
	       "Rate ok = %d\n"
	       "Rate violated = %d\nSS ok = %d\nSS over = %d\n"
	       "Nofeed = %d\n"
	       "Mo md drops = %d\n"
	       "No port drops = %d\n"
	       "Uncaught exceptions = %d\n"
	       "Init sent = %d\n",
	       tfc_stats.sendtotal, tfc_stats.acktotal, tfc_stats.nacktotal,
	       tfc_stats.badnonce,
	       tfc_stats.recvtotal, tfc_stats.input_drops, tfc_stats.acksent,
	       tfc_stats.sb_overflows,
	       tfc_stats.rate_ok, tfc_stats.rate_over, tfc_stats.ss_ok,
	       tfc_stats.ss_over, tfc_stats.nofeed,
	       tfc_stats.nodmdrops,
	       tfc_stats.noport_drops,
	       tfc_stats.uncaught,
	       tfc_stats.initsent
	       );
	       
	tfc_stats.sendtotal = tfc_stats.acktotal = 0;
	tfc_stats.nacktotal = tfc_stats.badnonce = 0;
	tfc_stats.recvtotal = tfc_stats.input_drops = 0;
	tfc_stats.acksent = tfc_stats.sb_overflows = 0;
	tfc_stats.rate_ok = tfc_stats.rate_over = tfc_stats.ss_ok = 0;
	tfc_stats.ss_over = 0;
	tfc_stats.uncaught = 0;
	tfc_stats.initsent = 0;
	
	return error;
}

/*
 * Initiate (or continue) disconnect.
 * If in ``let data drain'' option and linger null, just drop.
 * Otherwise (hard), mark socket disconnecting and drop
 * current input data;
 */
static int
tfc_detach(register struct tfc_cb *tp)
{
	int error = 0;
	int status = 0;
	struct socket *so = tp->inpcb->inp_socket;

	if ((so->so_options & SO_LINGER) && so->so_linger == 0)
		error = tfc_drop(tp, 0);
	else {
		soisdisconnecting(so);
		sbflush(&so->so_rcv);
		/* Tell the user protocol to disconnect quickly */
		status = tfi_upcall(tp->psw->pr_usrreqs->pru_detach, so, tp->cyc_psw);

	}
	if (status == TFC_DISCONNECT)
		tfc_close(tp);
	else
		/* Set a timeout, so that we can free the state
		   if user protocol doesn't.
		*/
		callout_reset(tp->tt_2msl, tfc_2msl, tfc_timer_2msl, tp);

	return error;
}

int
tfc_drop(struct tfc_cb *tp, int errno)
{
	struct socket *so = tp->inpcb->inp_socket;

	tfc_stats.tcps_conndrops++;
	if (errno == ETIMEDOUT && tp->t_softerror)
		errno = tp->t_softerror;
	so->so_error = errno;
	tfc_close(tp);
	return 0;

}
/* The connection should be completely terminated and the protocol should not be
   reachable from all directions */

void
tfc_timer_2msl(void *xtp)
{
	int s = splnet();
	struct tfc_cb *tp = xtp;
	if (callout_pending(tp->tt_2msl) || !callout_active(tp->tt_2msl)) {
		splx(s);
		return;
	}
	callout_deactivate(tp->tt_2msl);
	
	/* callout_stop for all timeouts */
	tfc_close(tp);
	splx(s);
}

void
tfc_close(struct tfc_cb *tp)
{
	struct mbufq_ent *q;
	struct inpcb *inp = tp->inpcb;
	struct socket *so = inp->inp_socket;
#ifdef TFC_LATER
	/* Save some state in the routing table entry to benefit
	   new connections */ 
	register struct rtentry *rt;
	int dosavessthresh;
#endif
	callout_stop(tp->tt_2msl);
	callout_stop(tp->tt_nofeed);
	/*
	 * XXX: Make sure that all of user timers are stopped before we
	 * delete the PCB.
	 */

	/* free any mbufs held by the protocol */
	while((q = LIST_FIRST(&tp->mbufq)) != NULL) {
		LIST_REMOVE(q, next);
		m_freem(q->m);
		FREE(q, M_TFCMBUFQ);
	}

	FREE(tp->scoreboard, M_TFCPROTO);
	tp->scoreboard = NULL;

	inp->inp_ppcb = NULL;
	soisdisconnected(so);

	printf("Detaching inpcb in tfc_close\n");
	in_pcbdetach(inp);
	printf("ok\n");
	if( tp->xp != NULL ) {
		tp->xp->ref_count -= 1;
	}
	tfc_stats.tcps_closed++;
}

#define	COMMON_START()	do { \
			        if (inp == 0) { \
				     splx(s); \
				     return EINVAL; \
				} \
				tp = intotfccb(inp); \
			        if (tp == 0 || tp->state < TFCS_ATTACHED) { \
				     splx(s); \
				     return EINVAL; \
				} \
		        } while(0)

#define COMMON_END(req)	out: splx(s); return error; goto out


/* If no protocol is attached, attach the default protocol to this socket */ 
#define ATTACH_DEFAULT() \
do { \
	if (inp == 0) { \
		splx(s); \
		return EINVAL; \
	} \
	tp = intotfccb(inp); \
	if (tp == 0) { \
		splx(s); \
		return EINVAL; \
	}  \
	if (tp->state < TFCS_ATTACHED) { \
		if (tp->dm->default_protop == NULL) { \
			error = TFC_NOPROTO;  \
		} else { \
			error = tfc_soattachproto(so, tp->dm->proto, \
						  tp->dm->default_protop, \
						  p); \
		} \
		if (error) \
			goto out; \
	} \
} while(0)

/*
 * Let the user protocol handle this anyway it wants
 */
static int
tfc_usr_disconnect(struct socket *so)
{
	int s = splnet();
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tfc_cb *tp;

	COMMON_START();
	error = tfi_upcall(tp->psw->pr_usrreqs->pru_disconnect, so, tp->cyc_psw);
	if (error)
		goto out;
	COMMON_END(PRU_DISCONNECT);
}

/*
 * Give the socket an address.
 */
/*
 * User protocol is not invoked
 */
static int
tfc_usr_bind(struct socket *so, struct sockaddr *nam, struct proc *p)
{
	int s = splnet();
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tfc_cb *tp;
	struct sockaddr_in *sinp;

	/* COMMON_START(); */
	
	/* if the socket is unattched, attach to default protocol. */
	ATTACH_DEFAULT();
	
	/*
	 * Must check for multicast addresses and disallow binding
	 * to them.
	 */
	sinp = (struct sockaddr_in *)nam;
	if (sinp->sin_family == AF_INET &&
	    IN_MULTICAST(ntohl(sinp->sin_addr.s_addr))) {
		error = EAFNOSUPPORT;
		goto out;
	}
	error = in_pcbbind(inp, nam, p);
	if (error)
		goto out;
	COMMON_END(PRU_BIND);
}

/*
 * User protocol does the main work
 */
static int
tfc_usr_connect(struct socket *so, struct sockaddr *nam, struct proc *p)
{
	int s = splnet();
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tfc_cb *tp;
	struct sockaddr_in *sinp;

	/* COMMON_START(); */
	
	/* if the socket is unattched, attach to default protocol. */
	ATTACH_DEFAULT();
	
	/*
	 * Must disallow ``connections'' to multicast addresses.
	 */
	sinp = (struct sockaddr_in *)nam;
	if (sinp->sin_family == AF_INET
	    && IN_MULTICAST(ntohl(sinp->sin_addr.s_addr))) {
		error = EAFNOSUPPORT;
		goto out;
	}

	/* Should check here if the local address is in use */
	
	if (inp->inp_laddr.s_addr == INADDR_ANY && p->p_prison != NULL)
		error = in_pcbbind(inp, NULL, p);
	if (error)
		goto out;
	prison_remote_ip(p, 0, &sinp->sin_addr.s_addr);
#ifdef VERBOSE_DEBUG
	printf("connect: upcall\n");
#endif

	error = tfi_upcall(tp->psw->pr_usrreqs->pru_connect, so, nam, p, tp->cyc_psw);
#ifdef VERBOSE_DEBUG
	printf("connect: result %d\n", error);
#endif
	if (error)
		goto out;
	COMMON_END(PRU_CONNECT);
}

/*
 * user protocol is given a chance to cleanup
 */
static int
tfc_usr_abort(struct socket *so)
{
	int s = splnet();
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tfc_cb *tp;

	COMMON_START();
	/* let the user protocol do some cleanup */
	tfi_upcall(tp->psw->pr_usrreqs->pru_abort, so, tp->cyc_psw);
	/* Drop this xtcb anyway */
	error = tfc_drop(tp, ECONNABORTED);

	COMMON_END(PRU_ABORT);
}

/*
 * Accept a connection.  Essentially all the work is
 * done at higher levels; just return the address
 * of the peer, storing through addr.
 */
/*
 * Not passed to the user protocol
 */
static int
tfc_usr_accept(struct socket *so, struct sockaddr **nam)
{
	int s = splnet();
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tcpcb *tp = NULL;

	if (so->so_state & SS_ISDISCONNECTED) {
		error = ECONNABORTED;
		goto out;
	}
	if (inp == 0) {
		splx(s);
		return (EINVAL);
	}
	tp = intotcpcb(inp);
	in_setpeeraddr(so, nam);
	COMMON_END(PRU_ACCEPT);
}

/*
 * Prepare to accept connections.
 */
/*
 * User protocol is notified
 */
static int
tfc_usr_listen(struct socket *so, struct proc *p)
{
	int s = splnet();
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tfc_cb *tp;

	/* COMMON_START(); */
	
	/* if the socket is unattched, attach to default protocol. */
	ATTACH_DEFAULT();
	
	if (inp->inp_lport == 0)
		error = in_pcbbind(inp, (struct sockaddr *)0, p);
	if (error)
		goto out;
	
	error = tfi_upcall(tp->psw->pr_usrreqs->pru_listen, so, p, tp->cyc_psw);
	if (error)
		goto out;
	
	COMMON_END(PRU_LISTEN);
}

/*
 * After a receive, possibly send window update to peer.
 */
static int
tfc_usr_rcvd(struct socket *so, int flags)
{
	int s = splnet();
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tfc_cb *tp;

	COMMON_START();

	error = tfi_upcall(tp->psw->pr_usrreqs->pru_rcvd, so, flags, tp->cyc_psw);
	if (error)
		goto out;

	COMMON_END(PRU_RCVD);
}

/*
 * Receive out-of-band data.
 */
static int
tfc_usr_rcvoob(struct socket *so, struct mbuf *m, int flags)
{
	int s = splnet();
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tfc_cb *tp;

	COMMON_START();
	if ((so->so_oobmark == 0 &&
	     (so->so_state & SS_RCVATMARK) == 0) ||
	    so->so_options & SO_OOBINLINE ) {
		error = EINVAL;
		goto out;
	}
	error = tfi_upcall(tp->psw->pr_usrreqs->pru_rcvoob, so, m, flags, tp->cyc_psw);
	if (error)
		goto out;

	COMMON_END(PRU_RCVOOB);
}


/*
 * Do a send by putting data in output queue and updating urgent
 * marker if URG set.  Possibly send more data.  Unlike the other
 * pru_*() routines, the mbuf chains are our responsibility.  We
 * must either enqueue them or free them.  The other pru_* routines
 * generally are caller-frees.
 */
/*
 * TFC will handle the flags and copy the mbuf into socket buffer.
 * Then the user protocol will be called to send out any data
 */
/*
 * TFC should not queue any data in the socket and maybe let the protocol
 * handle that.
 */
static int
tfc_usr_send(struct socket *so, int flags, struct mbuf *m, 
	     struct sockaddr *nam, struct mbuf *control, struct proc *p)
{
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tfc_cb *tp;
#ifdef PARANOIA
	int s = splnet();
#endif

	if (inp == NULL) {
		/*
		 * OOPS! we lost a race, the TCP session got reset after
		 * we checked SS_CANTSENDMORE, eg: while doing uiomove or a
		 * network interrupt in the non-splnet() section of sosend().
		 */
		error = ECONNRESET;	/* XXX EPIPE? */
		goto release;
	}
	
	if (nam) {
		struct sockaddr_in *sin = (struct sockaddr_in *)nam;
		prison_remote_ip(p, 0, &sin->sin_addr.s_addr);
	} else {
		if (inp->inp_faddr.s_addr == INADDR_ANY) {
			error = ENOTCONN;
			tfc_stats.sendnoconn++;
			goto release;
		}
	}
	
	tp = intotfccb(inp);
	/* Should we do this at splnet()??? */

	/* User protocol should free the mbuf m, the control mbuf is freed by TFC */
	error = tfi_upcall(tp->psw->pr_usrreqs->pru_send, so, flags, m, nam, control, p, tp->cyc_psw);
	goto out;

 release:
	if (m)
		m_freem(m);
 out:
#ifdef PARANOIA
	splx(s);
#endif
	if (control)
		m_freem(control);
	return error;
}

/*
 * Mark the connection as being incapable of further output.
 */
static int
tfc_usr_shutdown(struct socket *so)
{
	int s = splnet();
	int error = 0;
	struct inpcb *inp = sotoinpcb(so);
	struct tfc_cb *tp;

	COMMON_START();
	socantsendmore(so);
	error = tfi_upcall(tp->psw->pr_usrreqs->pru_shutdown, so, tp->cyc_psw);
	if (error)
		goto out;

	COMMON_END(PRU_SHUTDOWN);
}
