/*
 * tfci.h
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#ifndef TFCI_H
#define TFCI_H

/*
 * The tfi_proto structure should hold data pertaining to a protocol that
 * is static in nature, i.e., data that exists even when there is no connection
 * open for this protocol.  
 */
struct tfc_proto {
	SLIST_ENTRY(tfc_proto) next;
	u_long hash;
	struct tfc_protodomain *dm;
	struct tfi_protosw *psw;
	struct tfi_protosw_cyc *cyc_psw;
	u_long ref_count;       
	/* how many connections are currently using this protocol */ 
};

/*
 * Generic structure used by the protocols to extract port numbers from network
 * packets
 */
struct tfc_tp {
	u_short th_sport;
	u_short th_dport;
};

/* Generic transport header extractor */
int tfc_check_tphdr __P((struct mbuf **m, struct tfc_tp *tp, int off0));

SLIST_HEAD(tfc_protohead, tfc_proto);

/*
 * tfi_protodomain holds data pertaining to a particular domain of tfc protocols.
 * For example, TCP, DCCP, and TFRC will exist in their separate domains.
 */
struct tfc_protodomain {
	u_long hashsize; /* Size of the hashtable to store protocols */
	u_long hashmask; /* size of the hash table */
	u_short proto;
	struct inpcbhead xtcb; 
	struct inpcbinfo xtcbinfo; /* structures to hold all active protocols
				  in this domain */
	int (*check_hdr) __P((struct mbuf **, struct tfc_tp *, int));
	struct tfc_protohead *hashbase; /* hashtable of protocols */
	u_long default_proto_hash; /* hash of the default protocol */
	struct tfc_proto *default_protop; /* pointer to the default */
	u_char is_tcp; /* Whether the protocols in this domain are compatible
			  with TCP. Connection establishment and tear-down protocol.
		       */
};

/* Number of TFC domains */
#define TFC_NUM_DOMAINS 2
extern struct tfc_protodomain *tfc_domains[TFC_NUM_DOMAINS];

#define TFCCB_HASHTABLE_SIZE 512 /* size of the ports hashtable */

/* from code hash to index in the hashtable */
#define TFC_CHASH2INDEX(dm, hash) ((hash) & (dm)->hashmask)

/* Check the invariant on a protocol handle */
#define TFI_PROTO_INVARIANT(pr) (pr->p != NULL && pr->p->dm != NULL)

/* fn decls */
struct tfc_protodomain * protolookup_domain(u_short proto);
struct tfc_protodomain * hashlookup_domain(u_short domain);
int domain_addproto(struct tfc_protodomain *dm, u_long hash,
		    struct tfi_protosw *psw, struct tfi_protosw_cyc *cyc_psw, struct tfc_proto **xpp);
int domain_deleteproto(struct tfc_protodomain *d, struct tfc_proto *p);
struct tfc_proto *domain_findproto(struct tfc_protodomain *dm, u_long hash);
int tfc_soattachproto(struct socket* so, int proto, struct tfc_proto *xp,
		       struct proc *p);



extern volatile int tfi_inupcall, tfi_inkernel;
extern void tfi_timerstart(void);
extern int tfi_timerstop(void);

#define tfi_upcallenter()       tfi_inupcall++, tfi_timerstart()
#define tfi_upcallleave()       (void)tfi_timerstop(), tfi_inupcall--

/*#define NO_TFI_TIMER*/

/* Call into protocol */
#ifdef NO_TFI_TIMER
#define tfi_upcall(fn, args...)		(fn)(## args)
#define tfi_upcallv(fn, args...)	(fn)(## args)
#else
#define tfi_upcall(fn, args...) ({ \
	int _rv, _ink; \
	if ((_ink = tfi_inkernel) != 0) \
		tfi_downcallleave(); \
	tfi_upcallenter(); \
	_rv = (fn)(## args); \
	tfi_upcallleave(); \
	if (_ink) \
		tfi_downcallenter(); \
	_rv; \
})
#define tfi_upcallv(fn, args...) { \
	tfi_upcallenter(); \
	(fn)(## args); \
	tfi_upcallleave(); \
}
#endif

/* typedef int tick_t; */
/*#define get_tick() ticks *//* return kernel ticks */
/*
#define	TICK_LT(a,b)	((a) < (b))
#define	TICK_LEQ(a,b)   ((a) <= (b))
#define	TICK_GT(a,b)	((a) > (b)) 
#define	TICK_GEQ(a,b)   ((a) >= (b))
#define TICK_DIFF(a, b) ((a) - (b))
*/

#define tfc_tick_rate 1 /* microseconds */
typedef long long tick_t;

#define USEC2MSEC(t) ((t)/1000)

/*
#define TICK2MSEC(t) ((tick_t)(t)*1000/tsc_freq)
#define TICK2USEC(t) ((tick_t)(t)*1000000/tsc_freq)
*/
/* #define get_tick() rdtsc()*/ /* (rdtsc()/(tsc_freq/tfc_tick_rate)) */

#define MSEC2TIMER(t) (((t)*hz)/1000)

#define TICK2MSEC(t) ((t)/1000)
#define MSEC2TICK(t) ((t)*1000)
#define TICK2USEC(t) (t)
#define USEC2TICK(t) (t)
#define TICK2TIMER(t) (TICK2USEC(t)*hz/1000000)

#define get_tick(ts) do { \
                          struct timeval tv; \
                          microtime(&tv);    \
		          ts = ((long long)tv.tv_sec*1000000+tv.tv_usec); \
                   } while(0)

#define	TICK_LT(a,b)	((a) < (b))
#define	TICK_LEQ(a,b)   ((a) <= (b))
#define	TICK_GT(a,b)	((a) > (b)) 
#define	TICK_GEQ(a,b)   ((a) >= (b))
#define TICK_DIFF(a, b) ((a) - (b))
#define TICK_ADD(a, b)  ((a) + (b))
/*
#define	TICK_LT(a,b)	((int64_t)((a)-(b)) < 0)
#define	TICK_LEQ(a,b)   ((int64_t)((a)-(b)) <= 0)
#define	TICK_GT(a,b)	((int64_t)((a)-(b)) > 0)
#define	TICK_GEQ(a,b)   ((int64_t)((a)-(b)) >= 0)
#define TICK_DIFF(a, b) ((a) - (b))
*/
#define TFC_DOMAIN_TCP 0
#define TFC_DOMAIN_END 1

/* Cyclone stuff from bin/genfiles/precore_c.h */
#include <machine/setjmp.h>
struct _RuntimeStack {
  int tag; /* 0 for an exception handler, 1 for a region handle */
  struct _RuntimeStack *next;
};

struct _handler_cons {
  struct _RuntimeStack s;
  jmp_buf handler;
};

extern void _push_handler(struct _handler_cons *);
extern void _pop_handler(void);
extern int setjmp(jmp_buf);
extern struct _xtunion_struct *_exn_thrown;

struct _xtunion_struct { char *tag; };

#define set_handler(body) \
{ \
	struct _handler_cons _handler; \
	int _status = 0, retval = 0; \
	\
	if (setjmp(_handler.handler)) \
		_status = 1; \
	if (_status) { \
		printf("Warning: Uncaught exception %s in %s\n", \
		       _exn_thrown->tag, \
		       __func__); \
		tfc_stats.uncaught++; \
		retval = TFC_EXCEPTION; \
	} else { \
		_push_handler(&_handler); \
		body; \
		_pop_handler(); \
	} \
	return retval; \
}

/* TFI_TIMER should be enabled only while testing cyclone code. C code
   will cause a crash if used with the timer code is enabled.
*/
#define NO_TFI_TIMER

/*
  #define TEST_XUDP
  #define TEST_XTCP
  #define TEST_XTCP_SACK
  #define TEST_XUDP_CYC
  #define TEST_XTCP_CYC
  #define TEST_XTCP_SACK_CYC
*/

#endif /* TFCI_H */
