/*
 * tfi_timer.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/proc.h>
#include <sys/jail.h>

#include <vm/vm_zone.h>

#include <net/if.h>
#include <net/route.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_pcb.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcpip.h>

#include <machine/cpu.h>
#include <machine/limits.h>

#include <netinet/xtcp/tfci.h>
#include <netinet/xtcp/tfi.h>
#include <netinet/xtcp/tfc_stats.h>
#include <netinet/xtcp/tfc_error.h>

#define TFI_MAXTICKS	2

volatile int tfi_inupcall;
volatile int tfi_inupcall_ticks;
volatile int tfi_inkernel;
volatile int tfi_cpuerror;

void tfi_hardclock(struct clockframe *frame);

#ifndef NO_TFI_TIMER
static struct _xtunion_struct cpuexn = { "CPUBlown" };
void throw(struct _xtunion_struct *);
#endif

/*
 * Record the time at which an upcall was entered.
 */
void
tfi_timerstart(void)
{
#ifndef NO_TFI_TIMER
	if (tfi_inupcall > 10)
		panic("tfi_timerstart fubar: inupcall=%d, ticks=%d",
		      tfi_inupcall, ticks);

	if (tfi_inupcall == 1)
		tfi_inupcall_ticks = ticks;
#endif
}

/*
 * Stop timing the current upcall.
 * Returns the number of ticks used by this upcall.
 * The return value could be used by the caller to keep a cumulative time.
 */
int
tfi_timerstop(void)
{
#ifndef NO_TFI_TIMER
	if (tfi_inupcall > 10)
		panic("tfi_timerstop fubar: inupcall=%d, ticks=%d",
		      tfi_inupcall, ticks);

	if (tfi_inupcall == 1)
		return ticks - tfi_inupcall_ticks;
#endif
	return 0;
}

static void
tfi_throwexception(void)
{
#ifndef NO_TFI_TIMER
	if (!tfi_inkernel) {
		tfi_inupcall = 0;
		throw(&cpuexn);
	} else
		tfi_cpuerror = 1;
#endif
}

/*
 * Called at hardclock interrupt to determine if the current upcall
 * has run too long.
 */
void
tfi_hardclock(struct clockframe *frame)
{
#ifndef NO_TFI_TIMER
	if (!tfi_inupcall)
		return;

	if (tfi_inupcall != 1)
		panic("tfi_hardclock fubar: inupcall=%d, ticks=%d",
		      tfi_inupcall, ticks);

	/*
	 * Determine how long the current process has been in its upcall.
	 * If timelimit has been exceeded tweak frame PC to force an exception.
	 */
	if (ticks - tfi_inupcall_ticks > TFI_MAXTICKS)
		CLKF_PC(frame) = (int) tfi_throwexception;
#endif
}

/*
 * Call out from cyclone back into TFI
 * Make a note that we have non-abortable code.
 */
void
tfi_downcallenter(void)
{
#ifndef NO_TFI_TIMER
	if (tfi_inkernel != 0)
		panic("tfi_downcallenter fubar: inupcall=%d, inkernel=%d",
		      tfi_inupcall, tfi_inkernel);

	tfi_inkernel++;
#endif
}

/*
 * Return from TFI back into cyclone.
 * See if we should be aborted.
 */
void
tfi_downcallleave(void)
{
#ifndef NO_TFI_TIMER
	if (tfi_inkernel != 1)
		panic("tfi_downcallleave fubar: inupcall=%d, inkernel=%d",
		      tfi_inupcall, tfi_inkernel);

	tfi_inkernel--;
	if (tfi_cpuerror) {
		tfi_cpuerror = 0;
		tfi_throwexception();
	}
#endif
}
