/*
 * xldev.c
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose with or without fee is hereby granted, provided
 * that the above copyright notice and this permission notice
 * appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE UNIVERSITY OF UTAH
 * DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
 * EVENT SHALL THE UNIVERSITY OF UTAH BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#include <assert.h>

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/ioccom.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/fcntl.h>
#include <sys/vnode.h>

#include <netinet/in.h>

#include <netinet/xtcp/xldev.h>

#define CDEV_IOCTL1         _IOR('C', 1, u_int)

MALLOC_DEFINE(M_XLD_REQ, "xld_request", "XLD request entry");
MALLOC_DEFINE(M_XLD_D, "xld_d", "XLD descriptor");

int xldev_open(dev_t dev, int flags, int otyp, struct proc *procp)
{
    struct xld_d *xd;
    int retval = 0;
    
    if( flags & FWRITE )
    {
	/* The device is read-only. */
	retval = EACCES;
    }
    else if( dev->si_drv1 != NULL )
    {
	/* The device can only be accessed by one process at a time. */
	retval = EBUSY;
    }
    else
    {
	/* Allocate a new descriptor and attach it to the device. */
	MALLOC(xd, struct xld_d *, sizeof(*xd), M_XLD_D, M_WAITOK | M_ZERO);
	dev->si_drv1 = xd;
	
	STAILQ_INIT(&xd->xd_requests);
    }
    return( retval );
}

int xldev_req_append_in(dev_t dev, struct in_addr *ia, u_long hash)
{
    struct xld_d *xd;
    int retval = 0;

    if( (ia == NULL) || (dev == NULL) )
    {
	retval = EINVAL;
    }
    else if( (xd = (struct xld_d *)dev->si_drv1) == NULL )
    {
	/* The device hasn't been opened yet. */
	retval = ENOTCONN;
    }
    else
    {
	struct xld_request *xr;
	char *ia_bits;
	
	ia_bits = (char *)&ia->s_addr;
	/* Allocate the request, ... */
	MALLOC(xr,
	       struct xld_request *,
	       sizeof(*xr),
	       M_XLD_REQ,
	       M_WAITOK|M_ZERO);
	/* write a reasonable string to the buffer, ... */
	xr->xr_data.length = sprintf(xr->xr_data.buffer,
				     "%u.%u.%u.%u/%lx\n",
				     ia_bits[0] & 0xff,
				     ia_bits[1] & 0xff,
				     ia_bits[2] & 0xff,
				     ia_bits[3] & 0xff,
				     hash);
	/* queue the request, and ... */
	STAILQ_INSERT_TAIL(&xd->xd_requests, xr, xr_link);
	/* wakeup the listener in xldev_read. */
	wakeup(xd);
    }
    return( retval );
}

int xldev_close(dev_t dev, int flag, int otyp, struct proc *procp)
{
    struct xld_request *xr;
    struct xld_d *xd;
    int retval = 0;

    xd = (struct xld_d *)dev->si_drv1;
    /* Clear descriptor so the device can be reopened, ... */
    dev->si_drv1 = NULL;
    /* free the request list, and ... */
    while( (xr = STAILQ_FIRST(&xd->xd_requests)) != NULL )
    {
	STAILQ_REMOVE(&xd->xd_requests, xr, xld_request, xr_link);
	FREE(xr, M_XLD_REQ);
    }
    /* free the descriptor. */
    FREE(xd, M_XLD_D);
    
    return( retval );
}

int
xldev_ioctl(dev_t dev, u_long cmd, caddr_t arg, int mode, struct proc *procp)
{
    int error = 0;

    printf("xldev_ioctl: dev_t=%d, cmd=%lx, arg=%p, mode=%x procp=%p\n",
	   dev2udev(dev), cmd, arg, mode, procp);

    switch(cmd) {
    case CDEV_IOCTL1:
	printf("you called xldev_ioctl CDEV_IOCTL1\n");
	break;
    default:
	printf("No such ioctl for me!\n");
	error = EINVAL;
	break;
    }
    return error;
}

int xldev_write(dev_t dev, struct uio *uio, int ioflag)
{
    /* XXX */
    return( EPERM );
}

int xldev_read(dev_t dev, struct uio *uio, int ioflag)
{
    struct xld_request *xr;
    struct xld_d *xd;
    int retval = 0;

    xd = (struct xld_d *)dev->si_drv1;
    while( uio->uio_resid > 0 )
    {
	if( (xr = STAILQ_FIRST(&xd->xd_requests)) != NULL )
	{
	    unsigned int amount;

	    /* Theres a request, figure out how much to copy, ... */
	    amount = min(uio->uio_resid,
			 xr->xr_data.length - xr->xr_data.index);
	    /* do the copy, ... */
	    uiomove(&xr->xr_data.buffer[xr->xr_data.index], amount, uio);
	    /* clear resid to break the while, and ... */
	    uio->uio_resid = 0; /* XXX We should really try to read more... */
	    /* see if the request was completely consumed. */
	    xr->xr_data.index += amount;
	    if( xr->xr_data.index == xr->xr_data.length )
	    {
		STAILQ_REMOVE(&xd->xd_requests, xr, xld_request, xr_link);
		FREE(xr, M_XLD_REQ);
	    }
	}
	else
	{
	    /* No requests and ... */
	    if( ioflag & IO_NDELAY )
	    {
		/* nonblocking read, so we return. */
		retval = EWOULDBLOCK;
	    }
	    else
	    {
		int rc;

		/* blocking read, so we wait for a wakeup or signal. */
		rc = tsleep(xd, PZERO | PCATCH, "xld", 0);
		switch( rc )
		{
		case 0:
		    /* Got a wakeup(9), recheck the request queue. */
		    break;
		case EWOULDBLOCK:
		    /* Can't happen... */
		    break;
		case EINTR:
		case ERESTART:
		    /*
		     * Got a signal, clear resid to break the while and
		     * return.
		     */
		    uio->uio_resid = 0;
		    retval = rc;
		    break;
		}
	    }
	}
    }
    return( retval );
}
