/*
 * Copyright (c) 1982, 1986, 1988, 1990, 1993, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)udp_usrreq.c	8.6 (Berkeley) 5/23/95
 * $FreeBSD: src/sys/netinet/udp_usrreq.c,v 1.64.2.16 2002/08/07 16:14:47 luigi Exp $
 */

#include <netinet/xtcp/tfi.h>

#include <netinet/xtcp/xudp.h>
#include <netinet/xtcp/xudp_var.h>

/*
 * UDP protocol implementation.
 * Per RFC 768, August, 1980.
 */

/* globals */
extern struct pr_usrreqs udp_usrreqs;
#define TOS2NONCE(tos) ((tos) & 0x03);

/* PP: test the range API */
#define RANGE

/* locals */
/*
#ifndef	COMPAT_42
static int	udpcksum = 1;
#else
*/
static int	udpcksum = 0;		/* XXX */
/* #endif */

#ifdef XUDP_LATER
SYSCTL_INT(_net_inet_udp, UDPCTL_CHECKSUM, checksum, CTLFLAG_RW,
		&udpcksum, 0, "");
#endif

int	x_log_in_vain = 0;
#ifdef XUDP_LATER
SYSCTL_INT(_net_inet_udp, OID_AUTO, log_in_vain, CTLFLAG_RW, 
    &log_in_vain, 0, "Log all incoming UDP packets");
#endif

#ifdef XUDP_LATER
static int	blackhole = 0;
#endif

#ifdef XUDP_LATER
SYSCTL_INT(_net_inet_udp, OID_AUTO, blackhole, CTLFLAG_RW,
	&blackhole, 0, "Do not send port unreachables for refused connects");
#endif

struct	xudpstat xudpstat;	/* from udp_var.h */
#ifdef XUDP_LATER
SYSCTL_STRUCT(_net_inet_udp, UDPCTL_STATS, stats, CTLFLAG_RW,
    &udpstat, udpstat, "UDP statistics (struct udpstat, netinet/udp_var.h)");
#endif

static struct	sockaddr_in xudp_in = { sizeof(xudp_in), AF_INET };

#ifdef XUDP_LATER
static void xudp_append __P((struct inpcb *last, struct ip *ip,
			    struct mbuf *n, int off));
#endif

static int xudp_detach __P((struct socket *so, struct tfi_protosw_cyc *nu));
static	int xudp_output __P((struct inpcb *, struct mbuf *, struct sockaddr *,
			    struct mbuf *, struct proc *));
#ifdef RANGE
static int xudp_sendack __P((seq_t bseq, seq_t seq, u_char nonce, struct socket *so));
#else
static int xudp_sendack __P((seq_t seq, u_char nonce, struct socket *so));
#endif
/* All these should go into a pcb */
static seq_t seq; /* This xudp supports only one connection for now, no pcb */
static u_char nonce_sum;

/*
 * IP options have been stripped off from the mbuf.
 */

int
xudp_input(register struct mbuf *m, int off, int proto, struct socket *so, struct tfi_protosw_cyc *nu)
{
	int iphlen = off;
	register struct ip *ip;
	register struct xudphdr *uh;
#ifdef XUDP_LATER
	struct mbuf *opts = 0;
#endif
	int len;
	struct ip save_ip;
	struct sockaddr *append_sa;
	int error = 0;
	u_char pkt_type;
	u_char nonce;
	seq_t seq, bseq;
	
	xudpstat.udps_ipackets++;

	/*
	 * Since IP options have been stripped off, ip hlen is
	 * straight-forward
	 */
	iphlen = sizeof(struct ip);

	/*
	 * Get IP and UDP header together in first mbuf.
	 */
	if (m->m_len < iphlen + sizeof(struct xudphdr)) {
		if ((m = tfi_m_pullup(m, iphlen + sizeof(struct xudphdr))) == 0) {
			xudpstat.udps_hdrops++;
			printf ("%s:No buffer\n", __FILE__);
			return TFC_EINVAL;
		}
	}
	ip = tfi_mtod(m, struct ip *);
	/*
	 * XXX: Such typecasting might be problematic when we
	 * move to Cyclone.
	 * How can a protocol safely typecast a byte-array into a struct in
	 * Cyclone? Wilson had a paper on this thing.
	 */
	/*
	 * One simple way could be to generate code at compile time which redefines
	 * the type of an mbuf to be:
	   struct mbuf {
	     struct iphdr ip;
	     struct xudphdr udp;
	     char *data;
	   }
	   Only thing we need to ensure is that none of the fields in "xudphdr" is
	   a pointer. We'll be fine dealing with just bits.
	   
	 * The above seems like one workable solution for now. This is possible
	 * because the IP options have been stripped off.
	 * The IP options should be made available to application by the TFC layer
	 * and not the transport protocol itself. If the transport wants, it can get
	 * access to the IP options.
	 */
	uh = (struct xudphdr *)((caddr_t)ip + iphlen);

	/* destination port of 0 is illegal, based on RFC768. */
	if (uh->uh_dport == 0) {
		error = 1;
		goto bad;
	}

	/*
	 * Make mbuf data length reflect UDP length.
	 * If not enough data to reflect UDP length, drop.
	 */
	len = ntohs((u_short)uh->uh_ulen);
	if (ip->ip_len != len) {
		if (len > ip->ip_len || len < sizeof(struct xudphdr)) {
			xudpstat.udps_badlen++;
			error = 2;
			goto bad;
		}
		tfi_m_adj(m, len - ip->ip_len);
		/* ip->ip_len = len; */
	}
	/*
	 * Save a copy of the IP header in case we want restore it
	 * for sending an ICMP error message in response.
	 */
	save_ip = *ip;

	/*
	 * Checksum extended UDP header and data.
	 */
	if (uh->uh_sum) {
		if (m->m_pkthdr.csum_flags & CSUM_DATA_VALID) {
			if (m->m_pkthdr.csum_flags & CSUM_PSEUDO_HDR)
				uh->uh_sum = m->m_pkthdr.csum_data;
			else
	                	uh->uh_sum = in_pseudo(ip->ip_src.s_addr,
				    ip->ip_dst.s_addr, htonl((u_short)len +
				    m->m_pkthdr.csum_data + IPPROTO_TFC));
			uh->uh_sum ^= 0xffff;
		} else {
			char b[9];
			bcopy(((struct ipovly *)ip)->ih_x1, b, 9);
			bzero(((struct ipovly *)ip)->ih_x1, 9);
			((struct ipovly *)ip)->ih_len = uh->uh_ulen;
			uh->uh_sum = in_cksum(m, len + sizeof (struct ip));
			bcopy(b, ((struct ipovly *)ip)->ih_x1, 9);
		}
		if (uh->uh_sum) {
			xudpstat.udps_badsum++;
			error = 3;
			goto bad;
		}
	} else
		xudpstat.udps_nosum++;

#ifdef XUDP_LATER	
	if (IN_MULTICAST(ntohl(ip->ip_dst.s_addr)) ||
	    in_broadcast(ip->ip_dst, m->m_pkthdr.rcvif)) {
		struct inpcb *last;
		/*
		 * Deliver a multicast or broadcast datagram to *all* sockets
		 * for which the local and remote addresses and ports match
		 * those of the incoming datagram.  This allows more than
		 * one process to receive multi/broadcasts on the same port.
		 * (This really ought to be done for unicast datagrams as
		 * well, but that would cause problems with existing
		 * applications that open both address-specific sockets and
		 * a wildcard socket listening to the same port -- they would
		 * end up receiving duplicates of every unicast datagram.
		 * Those applications open the multiple sockets to overcome an
		 * inadequacy of the UDP socket interface, but for backwards
		 * compatibility we avoid the problem here rather than
		 * fixing the interface.  Maybe 4.5BSD will remedy this?)
		 */

		/*
		 * Construct sockaddr format source address.
		 */
		xudp_in.sin_port = uh->uh_sport;
		xudp_in.sin_addr = ip->ip_src;
		/*
		 * Locate pcb(s) for datagram.
		 * (Algorithm copied from raw_intr().)
		 */
		last = NULL;
		LIST_FOREACH(inp, &udb, inp_list) {
			if (inp->inp_lport != uh->uh_dport)
				continue;
			if (inp->inp_laddr.s_addr != INADDR_ANY) {
				if (inp->inp_laddr.s_addr !=
				    ip->ip_dst.s_addr)
					continue;
			}
			if (inp->inp_faddr.s_addr != INADDR_ANY) {
				if (inp->inp_faddr.s_addr !=
				    ip->ip_src.s_addr ||
				    inp->inp_fport != uh->uh_sport)
					continue;
			}

			if (last != NULL) {
				struct mbuf *n;
				if ((n = m_copy(m, 0, M_COPYALL)) != NULL)
					udp_append(last, ip, n,
						   iphlen +
						   sizeof(struct xudphdr));
			}
			last = inp;
			/*
			 * Don't look for additional matches if this one does
			 * not have either the SO_REUSEPORT or SO_REUSEADDR
			 * socket options set.  This heuristic avoids searching
			 * through all pcbs in the common case of a non-shared
			 * port.  It * assumes that an application will never
			 * clear these options after setting them.
			 */
			if ((last->inp_socket->so_options&(SO_REUSEPORT|SO_REUSEADDR)) == 0)
				break;
		}

		if (last == NULL) {
			/*
			 * No matching pcb found; discard datagram.
			 * (No need to send an ICMP Port Unreachable
			 * for a broadcast or multicast datgram.)
			 */
			xudpstat.udps_noportbcast++;
			error = 4;
			goto bad;
		}
		udp_append(last, ip, m, iphlen + sizeof(struct xudphdr));
		return;
	}
#endif /* TFC_LATER */

	/*
	 * Locate pcb for datagram.
	 */
	/*
	 * PP: A PCB has already been located for this datagram.
	 */

	/*
	 * Construct sockaddr format source address.
	 * Stuff source address and datagram in user buffer.
	 */
	xudp_in.sin_port = uh->uh_sport;
	xudp_in.sin_addr = ip->ip_src;

	/*
	 * PP: According to a comment in the book, TCP/IP Illus. Vol. 2, pg 776,
	 * this option doesn't make
	 * sense anymore. So maybe we'll just get rid of it.
	 */
#ifdef XUDP_LATER
	if (inp->inp_flags & INP_CONTROLOPTS
	    || inp->inp_socket->so_options & SO_TIMESTAMP) {
		ip_savecontrol(inp, &opts, ip, m);
	}
#endif
	pkt_type = uh->uh_type;

	switch(pkt_type) {
	case XUDP_DATA:
		printf("xudp data packet\n");
		seq = ntohl(uh->uh_seq);
#ifdef RANGE
		bseq = ntohl(uh->uh_bseq);
#endif
		nonce = TOS2NONCE(ip->ip_tos);
		tfi_m_adj(m, iphlen + sizeof(struct xudphdr));
		append_sa = (struct sockaddr *)&xudp_in;
		
		/*
		  if (tfi_sbappendaddr(&inp->inp_socket->so_rcv, append_sa, m, opts) == 0)
		  {
		*/
		error = tfi_sbrappend(so, m);
		/* send an ack for this packet */
		{
			int te;
			/* printf("xudp_input: recvd seq %d, nonce %d\n", seq, nonce); */
#ifdef RANGE
			te = xudp_sendack(bseq, seq, nonce, so);
#else
			te = xudp_sendack(seq, nonce, so);
#endif
			if (te)
				printf("xudp_sendack failed: %d\n", te);
		}
		if (error) {
			/* printf ("%s: Error while queueing data\n", __FILE__); */
			goto bad;
		}

		tfi_sorwakeup(so);
		
		/* ignore error from sendack, or tfc may free the mbuf that is
		   queued at the socket
		*/
		break;
	case XUDP_ACK:
		/* ack packet */
		seq = ntohl(uh->uh_seq);
		nonce = uh->uh_noncesum;
#ifdef RANGE
		bseq = ntohl(uh->uh_bseq);
		error = tfi_acksum_range(bseq, seq, nonce, so);
#else
		error = tfi_acksum(seq, nonce, so);
#endif
		/* Free the packet and ignore any error for now */
		tfi_m_freem(m);
		if (error) {
			/* printf ("Ack error %d for seq %d\n", error, seq); */
			error = 0;
		}
		break;
	case XUDP_CE:
		/* nack this packet */
		seq = ntohl(uh->uh_seq);
		error = tfi_nack(seq, so);
		/* Free the packet and ignore any error for now */
		tfi_m_freem(m);
		if (error) {
			printf ("Nack error %d for seq %d\n", error, seq);
			error = 0;
		}
		break;
	default:
		printf("%s: Unrecognized packet type\n", __FILE__);
	}
	if (error)
		goto bad;
	
	return 0;
 bad:
	/* The mbuf should be freed by the TFC layer so that there is
	   no confusion as to who owns it.
	m_freem(m);
	*/
	printf ("XUDP receive error = %d\n", error);
	return error;
}
/*
 * send out an ack that has no data in it
 */
static int
#ifdef RANGE
xudp_sendack(seq_t bseq, seq_t seq, u_char nonce, struct socket *so)
#else
xudp_sendack(seq_t seq, u_char nonce, struct socket *so)
#endif
{
	struct mbuf *nm;
	struct xudpiphdr *nui;
	struct inpcb *inp = sotoinpcb(so);
	int hdrlen = sizeof(struct xudpiphdr);
	int len = 0;
	int error = 0;
	
	if (inp == NULL)
		return EINVAL;

	if (nonce == 0) {
		/* no need to send an ack for this one */
		return 0;
	}
	
	nm = tfi_mgethdr();
	if (nm == NULL) {
		printf("unable to get mbuf for ack packet\n");
		return ENOBUFS;
	}

	/* We add XUDP and IP headers to the mbuf. */
	error = tfi_msetlen(nm, hdrlen); 
	if (error)
		goto release;
	error = tfi_msethlen(nm, hdrlen); 
	if (error)
		goto release;
	
	nui = tfi_mtod(nm, struct xudpiphdr *);
	
	bzero(nui->ui_x1, sizeof(nui->ui_x1));	/* XXX still needed? */
	nui->ui_pr = IPPROTO_TFC; /* XXX: */
	nui->ui_src = inp->inp_laddr;
	nui->ui_dst = inp->inp_faddr;
	nui->ui_sport = inp->inp_lport;
	nui->ui_dport = inp->inp_fport;
	nui->ui_ulen = htons(sizeof(struct xudphdr));

	/*
	 * Set up checksum and output datagram.
	 */
	if (udpcksum) {
        	nui->ui_sum = in_pseudo(nui->ui_src.s_addr, nui->ui_dst.s_addr,
		    htons((u_short)len + sizeof(struct xudphdr) + IPPROTO_TFC));
		nm->m_pkthdr.csum_flags = CSUM_UDP;
		nm->m_pkthdr.csum_data = offsetof(struct xudphdr, uh_sum);
	} else {
		nui->ui_sum = 0;
	}
	((struct ip *)nui)->ip_len = sizeof (struct xudpiphdr);
	((struct ip *)nui)->ip_ttl = inp->inp_ip_ttl;
	((struct ip *)nui)->ip_tos = inp->inp_ip_tos;

	nui->ui_seq = htonl(seq);
#ifdef RANGE
	nui->ui_bseq = htonl(bseq);
#endif
	if (nonce == 3) {
		printf("CE data received\n");
		nui->ui_type = XUDP_CE;
	} else {
		nui->ui_type = XUDP_ACK;
		nonce_sum = NONCE_SUM(nonce_sum, nonce);
	}
	nui->ui_noncesum = nonce_sum;

	/* tfi_sendack will free the mbuf */
	error = tfi_sendack(nm, inp);
	goto out;
	
 release:
	tfi_m_freem(nm);
 out:
	return error;
}

/*
 * Not needed for now.
 */
#ifdef XUDP_LATER
/*
 * subroutine of udp_input(), mainly for source code readability.
 * caller must properly init udp_ip6 and udp_in6 beforehand.
 */
static void
xudp_append(last, ip, n, off)
	struct inpcb *last;
	struct ip *ip;
	struct mbuf *n;
	int off;
{
	struct sockaddr *append_sa;
	struct mbuf *opts = 0;

	if (last->inp_flags & INP_CONTROLOPTS ||
	    last->inp_socket->so_options & SO_TIMESTAMP) {
#ifdef INET6
		if (last->inp_vflag & INP_IPV6) {
			int savedflags;

			if (udp_ip6.uip6_init_done == 0) {
				ip_2_ip6_hdr(&udp_ip6.uip6_ip6, ip);
				udp_ip6.uip6_init_done = 1;
			}
			savedflags = last->inp_flags;
			last->inp_flags &= ~INP_UNMAPPABLEOPTS;
			ip6_savecontrol(last, &opts, &udp_ip6.uip6_ip6, n);
			last->inp_flags = savedflags;
		} else
#endif
		ip_savecontrol(last, &opts, ip, n);
	}
#ifdef INET6
	if (last->inp_vflag & INP_IPV6) {
		if (udp_in6.uin6_init_done == 0) {
			in6_sin_2_v4mapsin6(&udp_in, &udp_in6.uin6_sin);
			udp_in6.uin6_init_done = 1;
		}
		append_sa = (struct sockaddr *)&udp_in6.uin6_sin;
	} else
#endif
	append_sa = (struct sockaddr *)&udp_in;
	m_adj(n, off);
	if (sbappendaddr(&last->inp_socket->so_rcv, append_sa, n, opts) == 0) {
		m_freem(n);
		if (opts)
			m_freem(opts);
		xudpstat.udps_fullsock++;
	} else
		sorwakeup(last->inp_socket);
}

#endif /* TFC_LATER */

/*
 * Notify a udp user of an asynchronous error;
 * just wake up so that he can collect error status.
 */
void
xudp_notify(register struct socket *so, int errno, struct tfi_protosw_cyc *nu)
{
	tfi_soseterror(so, errno);
	tfi_sorwakeup(so);
	tfi_sowwakeup(so);
}

/*
 * Looks like only the notify function is required.
 * The decision to wakeup all or some processes should be made by the TFC layer.
 */
#ifdef XUDP_LATER
void
udp_ctlinput(int cmd, struct sockaddr *sa, void *vip)
{
	struct ip *ip = vip;
	struct xudphdr *uh;
	void (*notify) __P((struct socket *, int)) = udp_notify;
        struct in_addr faddr;
	struct inpcb *inp;
	int s;

	faddr = ((struct sockaddr_in *)sa)->sin_addr;
	if (sa->sa_family != AF_INET || faddr.s_addr == INADDR_ANY)
        	return;

	if (PRC_IS_REDIRECT(cmd)) {
		ip = 0;
		notify = in_rtchange;
	} else if (cmd == PRC_HOSTDEAD)
		ip = 0;
	else if ((unsigned)cmd >= PRC_NCMDS || inetctlerrmap[cmd] == 0)
		return;
	if (ip) {
		s = splnet();
		uh = (struct xudphdr *)((caddr_t)ip + (ip->ip_hl << 2));
		inp = in_pcblookup_hash(&udbinfo, faddr, uh->uh_dport,
                    ip->ip_src, uh->uh_sport, 0, NULL);
		if (inp != NULL && inp->inp_socket != NULL)
			(*notify)(inp, inetctlerrmap[cmd]);
		splx(s);
	} else
		in_pcbnotifyall(&udb, faddr, inetctlerrmap[cmd], notify);
}

static int
udp_pcblist(SYSCTL_HANDLER_ARGS)
{
	int error, i, n, s;
	struct inpcb *inp, **inp_list;
	inp_gen_t gencnt;
	struct xinpgen xig;

	/*
	 * The process of preparing the TCB list is too time-consuming and
	 * resource-intensive to repeat twice on every request.
	 */
	if (req->oldptr == 0) {
		n = udbinfo.ipi_count;
		req->oldidx = 2 * (sizeof xig)
			+ (n + n/8) * sizeof(struct xinpcb);
		return 0;
	}

	if (req->newptr != 0)
		return EPERM;

	/*
	 * OK, now we're committed to doing something.
	 */
	s = splnet();
	gencnt = udbinfo.ipi_gencnt;
	n = udbinfo.ipi_count;
	splx(s);

	xig.xig_len = sizeof xig;
	xig.xig_count = n;
	xig.xig_gen = gencnt;
	xig.xig_sogen = so_gencnt;
	error = SYSCTL_OUT(req, &xig, sizeof xig);
	if (error)
		return error;

	inp_list = malloc(n * sizeof *inp_list, M_TEMP, M_WAITOK);
	if (inp_list == 0)
		return ENOMEM;
	
	s = splnet();
	for (inp = LIST_FIRST(udbinfo.listhead), i = 0; inp && i < n;
	     inp = LIST_NEXT(inp, inp_list)) {
		if (inp->inp_gencnt <= gencnt && !prison_xinpcb(req->p, inp))
			inp_list[i++] = inp;
	}
	splx(s);
	n = i;

	error = 0;
	for (i = 0; i < n; i++) {
		inp = inp_list[i];
		if (inp->inp_gencnt <= gencnt) {
			struct xinpcb xi;
			xi.xi_len = sizeof xi;
			/* XXX should avoid extra copy */
			bcopy(inp, &xi.xi_inp, sizeof *inp);
			if (inp->inp_socket)
				sotoxsocket(inp->inp_socket, &xi.xi_socket);
			error = SYSCTL_OUT(req, &xi, sizeof xi);
		}
	}
	if (!error) {
		/*
		 * Give the user an updated idea of our state.
		 * If the generation differs from what we told
		 * her before, she knows that something happened
		 * while we were processing this request, and it
		 * might be necessary to retry.
		 */
		s = splnet();
		xig.xig_gen = udbinfo.ipi_gencnt;
		xig.xig_sogen = so_gencnt;
		xig.xig_count = udbinfo.ipi_count;
		splx(s);
		error = SYSCTL_OUT(req, &xig, sizeof xig);
	}
	free(inp_list, M_TEMP);
	return error;
}

SYSCTL_PROC(_net_inet_udp, UDPCTL_PCBLIST, pcblist, CTLFLAG_RD, 0, 0,
	    udp_pcblist, "S,xinpcb", "List of active UDP sockets");

static int
xudp_getcred(SYSCTL_HANDLER_ARGS)
{
	struct sockaddr_in addrs[2];
	struct inpcb *inp;
	int error, s;

	error = suser(req->p);
	if (error)
		return (error);
	error = SYSCTL_IN(req, addrs, sizeof(addrs));
	if (error)
		return (error);
	s = splnet();
	inp = in_pcblookup_hash(&udbinfo, addrs[1].sin_addr, addrs[1].sin_port,
				addrs[0].sin_addr, addrs[0].sin_port, 1, NULL);
	if (inp == NULL || inp->inp_socket == NULL) {
		error = ENOENT;
		goto out;
	}
	error = SYSCTL_OUT(req, inp->inp_socket->so_cred, sizeof(struct ucred));
out:
	splx(s);
	return (error);
}

SYSCTL_PROC(_net_inet_udp, OID_AUTO, getcred, CTLTYPE_OPAQUE|CTLFLAG_RW,
    0, 0, udp_getcred, "S,ucred", "Get the ucred of a UDP connection");

#endif /* TFC_LATER */


static int
xudp_output(inp, m, addr, control, p)
	register struct inpcb *inp;
	struct mbuf *m;
	struct sockaddr *addr;
	struct mbuf *control;
	struct proc *p;
{
	register struct xudpiphdr *ui;
	register int len = m->m_pkthdr.len;
	struct in_addr laddr;
	struct sockaddr_in *sin;
	int s = 0, error = 0;

	if (len + sizeof(struct xudpiphdr) > IP_MAXPACKET) {
		error = EMSGSIZE;
		goto release;
	}

	if (addr) {
		sin = (struct sockaddr_in *)addr;
		/* prison_remote_ip(p, 0, &sin->sin_addr.s_addr); */
		laddr = inp->inp_laddr;
		if (inp->inp_faddr.s_addr != INADDR_ANY) {
			error = EISCONN;
			goto release;
		}
		/*
		 * Must block input while temporarily connected.
		 */
		s = tfi_splnet();
		error = tfi_in_pcbconnect(inp, addr, p);
		tfi_splx(s);
		if (error) {
			goto release;
		}
	} else {
		if (inp->inp_faddr.s_addr == INADDR_ANY) {
			error = ENOTCONN;
			goto release;
		}
	}
	/*
	 * Calculate data length and get a mbuf
	 * for UDP and IP headers.
	 */
	
	/*** XXX: what should we do with this? ***/
	M_PREPEND(m, sizeof(struct xudpiphdr), M_DONTWAIT);
	if (m == 0) {
		printf("prepend error\n");
		error = ENOBUFS;
		goto release;
	}

	/*
	 * Fill in mbuf with extended UDP header
	 * and addresses and length put into network format.
	 */
	ui = tfi_mtod(m, struct xudpiphdr *);
	bzero(ui->ui_x1, sizeof(ui->ui_x1));	/* XXX still needed? */
	ui->ui_pr = IPPROTO_TFC; /* XXX: */
	ui->ui_src = inp->inp_laddr;
	ui->ui_dst = inp->inp_faddr;
	ui->ui_sport = inp->inp_lport;
	ui->ui_dport = inp->inp_fport;
	ui->ui_ulen = htons((u_short)len + sizeof(struct xudphdr));

	/*
	 * Set up checksum and output the datagram.
	 */
	if (udpcksum) {
        	ui->ui_sum = in_pseudo(ui->ui_src.s_addr, ui->ui_dst.s_addr,
		    htons((u_short)len + sizeof(struct xudphdr) + IPPROTO_TFC));
		m->m_pkthdr.csum_flags = CSUM_UDP;
		m->m_pkthdr.csum_data = offsetof(struct xudphdr, uh_sum);
	} else {
		ui->ui_sum = 0;
	}
	((struct ip *)ui)->ip_len = sizeof (struct xudpiphdr) + len;
	((struct ip *)ui)->ip_ttl = inp->inp_ip_ttl;	/* XXX */
	((struct ip *)ui)->ip_tos = inp->inp_ip_tos;	/* XXX */
	xudpstat.udps_opackets++;
	ui->ui_seq = htonl(++seq);
	ui->ui_type = XUDP_DATA;

	/* netsend will free the mbuf */
#ifdef RANGE
	ui->ui_bseq = htonl(seq);
	ui->ui_seq = htonl(seq+len);
	error = tfi_netsend_range(m, inp, seq, seq+len);
	seq += len;
#else
	error = tfi_netsend(m, inp, seq);
#endif	
	/* Fast send without no checking */
	/* error = tfi_sendack(m, inp); */
	goto out;

 release:
	tfi_m_freem(m);
 out:
	return (error);
}

u_long	xudp_sendspace = 9216;		/* really max datagram size */
					/* 40 1K datagrams */
#ifdef XUDP_LATER
SYSCTL_INT(_net_inet_udp, UDPCTL_MAXDGRAM, maxdgram, CTLFLAG_RW,
    &xudp_sendspace, 0, "Maximum outgoing UDP datagram size");
#endif

u_long	xudp_recvspace = 40 * (1024 + sizeof(struct sockaddr_in));
#ifdef XUDP_LATER
SYSCTL_INT(_net_inet_udp, UDPCTL_RECVSPACE, recvspace, CTLFLAG_RW,
    &xudp_recvspace, 0, "Maximum incoming UDP datagram size");
#endif

static int
xudp_abort(struct socket *so, struct tfi_protosw_cyc *nu)
{
	/* Everything is taken care of by the tfc layer */
	return 0;
}

/*
 * pru_attach will be called from the TFC layer after it has processed
 * the USR_PROTOCOL_ATTACH request from the socket options or from another
 * protocol.
 */
static int
xudp_attach(struct socket *so, int proto, struct proc *p, struct tfi_protosw_cyc *nu)
{
	int error = 0;

	error = tfi_soreserve(so, xudp_sendspace, xudp_recvspace);	
	if (error)
		return error;

	/* udp doesn't need a control block as yet.
	 * XXX: Add a control block later to keep seqno information.
	 */
	error = tfi_pcballoc(so, 32);
	if (error)
		return error;

	if (tfi_sotouserpcb(so) == NULL) {
		printf ("pcb not found\n");
		return EINVAL;
	}

	error = tfi_pcbdelete(so);

	return error;
}

/*
 * Bind is not needed to be processed by the protocol
 */
#ifdef TFC_IGNORE
static int
xudp_bind(struct socket *so, struct sockaddr *nam, struct proc *p)
{
	struct inpcb *inp;
	int s, error;

	inp = sotoinpcb(so);
	if (inp == 0)
		return EINVAL;
	s = splnet();
	error = in_pcbbind(inp, nam, p);
	splx(s);
	return error;
}
#endif

/*
 * The proc structure should be opaque for all protocol functions.
 */
/*
 * @splnet
 */
static int
xudp_connect(struct socket *so, struct sockaddr *nam, struct proc *p, struct tfi_protosw_cyc *nu)
{
	struct inpcb *inp;
	int error;

	inp = tfi_sotoinpcb(so);
	if (inp == 0) {
		return EINVAL;
	}
	
	if (inp->inp_faddr.s_addr != INADDR_ANY)
		return EISCONN;

	error = 0;
	
	if (inp->inp_lport == 0 && inp->inp_laddr.s_addr == INADDR_ANY)
		error = tfi_in_pcbbind(inp, NULL, p);

	if (error == 0) {
		error = tfi_in_pcbconnect(inp, nam, p);
		if (error)
			printf("**** connect error ****\n");
	}
	if (error == 0)
		tfi_soisconnected(so);
	return error;
}

/*
 * @splnet
 */
static int
xudp_detach(struct socket *so, struct tfi_protosw_cyc *nu)
{
	/* tell tfc to go ahead with disconnection */
	return TFC_DISCONNECT;
}

/*
 * @splnet
 */
static int
xudp_disconnect(struct socket *so, struct tfi_protosw_cyc *nu)
{
	struct inpcb *inp;

	inp = sotoinpcb(so);
	if (inp == 0)
		return EINVAL;
	if (inp->inp_faddr.s_addr == INADDR_ANY)
		return ENOTCONN;

	tfi_in_pcbdisconnect(inp);
	/* inp->inp_laddr.s_addr = INADDR_ANY; */
	/* so->so_state &= ~SS_ISCONNECTED; */
	tfi_sounsetstate(so, SS_ISCONNECTED);
	return 0;
}

/*
 * @normal
 */
static int
xudp_send(struct socket *so, int flags, struct mbuf *m, struct sockaddr *addr,
	    struct mbuf *control, struct proc *p, struct tfi_protosw_cyc *nu)
{
	struct inpcb *inp;

	inp = tfi_sotoinpcb(so);
	if (inp == 0) {
		tfi_m_freem(m);
		return EINVAL;
	}
	return xudp_output(inp, m, addr, control, p);
}

/*
 * @splnet
 */
int
xudp_shutdown(struct socket *so, struct tfi_protosw_cyc *nu)
{
	return 0;
}

/*
 * The notsupp entries should be filled in by the TFC layer
 */
struct tfi_usrreqs xudp_usrreqs = {
	xudp_abort, tfi_pru_accept_notsupp, xudp_attach, xudp_connect, 
	xudp_detach, xudp_disconnect, 
	tfi_pru_listen_notsupp, tfi_pru_rcvd_notsupp, 
	tfi_pru_rcvoob_notsupp, xudp_send, xudp_shutdown
};

struct tfi_protosw xudp_sw = { /* SOCK_STREAM, IPPROTO_TFC, 0, */
				xudp_input, 0, xudp_notify, 0,
				xudp_init, 0, 0, 0, &xudp_usrreqs };

/*
 * For UDP:
 * There is no need to maintain a list of udpcbS because it'll be taken
 * care of by the tfc layer.
 */

struct tfi_proto pr;

#define TFC_ATTACH_PROTOCOL 1
#define XUDP_HASH 169

int xudp_init(struct tfi_protosw_cyc *nu)
{
	/*
	 * Call tfc_add_proto to add this protocol in the TFC family.
	 */
	int error = 0;
	error = tfi_addproto(IPPROTO_TFC, XUDP_HASH, &xudp_sw, &pr);
	return error;
}

static int
xtcp_modevent(module_t mod, int type, void *data)
{
	int retval = 0;
	
	switch (type) {
	case MOD_LOAD:
		printf("\nXTCP protocol: UDP\n");
		retval = xudp_init(NULL);
		printf("xtcp_init(NULL) = %d\n", retval);
		break;

	case MOD_UNLOAD:
#if !defined(KLD_MODULE)
		printf("xtcp proto tcp statically compiled, cannot unload\n");
		return EINVAL ;
#else
		printf("tfi_delproto(%p)\n", &pr);
		retval = tfi_delproto(&pr);
		printf("%d\n", retval);
#endif
		break ;
	default:
		break ;
	}
	return retval;
}

static moduledata_t xtcp_proto_udp_mod = {
	"xtcp_proto_udp",
	xtcp_modevent,
	NULL
};
DECLARE_MODULE(xtcp_proto_udp, xtcp_proto_udp_mod, SI_SUB_PROTO_DOMAIN, SI_ORDER_ANY);
MODULE_VERSION(xtcp_proto_udp, 1);
